#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
class options_item:
  def __init__(self,value,description,dtype,name,permissible_values=None,value_description=None, descr_for_input_boxes=[],category='General'):
    """permissible values can be a vector or a string with an inequality, 
    where %s represents the number, for example "1>%s>0"\n
    if permissible_values is a vector, value_description is a corresponding vector with 
    description of each value in permissible_values"""
    #if permissible_values
    self.description=description
    self.value=value
    self.dtype=dtype
    if type(dtype)==str:
      self.dtype_str=dtype
    elif type(dtype)==list or type(dtype)==tuple:
      self.dtype_str=str(dtype).replace('<class ','').replace('[','').replace(']','').replace('>','').replace("'",'')
    else:
      self.dtype_str=dtype.__name__

    self.permissible_values=permissible_values
    self.value_description=value_description
    self.descr_for_input_boxes=descr_for_input_boxes
    self.category=category
    self.name=name
    self.selection_var= len(descr_for_input_boxes)==0 and type(permissible_values)==list
    self.is_inputlist=len(self.descr_for_input_boxes)>0



  def set(self,value,i=None):
    try:
      if not self.valid(value,i):
        return False
    except Exception as e:
      a=self.valid(value,i)
      print(e)
      return False
    if i is None:
      if str(self.value)!=str(value):
        self.value=value
    else:
      if self.value[i]!=value:
        self.value[i]=value
      else:
        return False
    return True

  def valid(self,value,i=None):
    if self.permissible_values is None:
      try:
        if self.dtype(value)==value:
          return True
      except:
        pass
      if type(value) in self.dtype:
        return True
    if i is None:
      return self.valid_test(value, self.permissible_values)
    else:
      return self.valid_test(value, self.permissible_values[i])

  def valid_test(self,value,permissible):
    if permissible is None:
      return True
    if type(permissible)==list or type(permissible)==tuple:
      try:
        if not type(value)==list or type(value)==tuple:
          value=self.dtype(value)
          return value in permissible
        else:
          valid=True
          for i in range(len(value)):
            value[i]=self.dtype(value[i])
            valid=valid*eval(permissible[i] %(value[i],))
      except:
        return False
      return valid
    elif type(permissible)==str:
      return eval(permissible %(value,))
    else:
      print('No method to handle this permissible')



class options():
  def __init__(self):
    pass


  def make_category_tree(self):
    opt=self.__dict__
    d=dict()
    keys=np.array(list(opt.keys()))
    keys=keys[keys.argsort()]
    for i in opt:
      if opt[i].category in d:
        d[opt[i].category].append(opt[i])
      else:
        d[opt[i].category]=[opt[i]]
      opt[i].code_name=i
    self.categories=d	
    keys=np.array(list(d.keys()))
    self.categories_srtd=keys[keys.argsort()]



def regression_options():
  #Add option here for it to apear in the "options"-tab. The options are bound
  #to the data sets loaded. Hence, a change in the options here only has effect
  #ON DATA SETS LOADED AFTER THE CHANGE
  self=options()
  self.accuracy					= options_item(0, 				"Accuracy of the optimization algorithm. 0 = fast and inaccurate, 3=slow and maximum accuracy", int, 
                                                                            'Accuracy', "%s>0",category='Regression')

  self.add_intercept				= options_item(True,			"If True, adds intercept if not all ready in the data",
                                                                         bool,'Add intercept', [True,False],['Add intercept','Do not add intercept'],category='Regression')
  self.arguments					= options_item(None, 				"A dict or string defining a dictionary in python syntax containing the initial arguments." 
                                                                             "An example can be obtained by printing ll.args.args_d"
                                                                                                                                        , [str,dict, list, np.ndarray], 'Initial arguments')	

  self.ARMA_constraint	        = options_item(1000.0,				'Maximum absolute value of ARMA coefficients', float, 'ARMA coefficient constraint',
                                                           None,None,category='ARIMA-GARCH')	
  self.constraints_engine	        = options_item(True,			'Determines whether to invoke the constraints engine', bool, 'Invoke constraints engine',
                                                              [True,False],['Invoke constraints','Do not invoke constraints'],category='Regression')	

  self.debug_mode	      	        = options_item(False,			'Determines whether the code will run in debug mode. Should normally allways be False', 
                                                                    bool, 'Debug or not',
                                                                                                                                        [True,False],['Debug mode','Not debug mode'],category='General')	

  self.multicoll_threshold_report = options_item(30,				'Threshold for reporting multicoll problems', float, 'Multicollinearity threshold',
                                                       None,None)		

  self.multicoll_threshold_max    = options_item(100,			'Threshold for imposing constraints on collineary variables', float, 'Multicollinearity threshold',
                                                       None,None)			

  self.parallel  					= options_item(True,			"If True, uses multiple cores for parallel computing",
                                                                              bool,'Multicore', [True,False],['Multicore','Single core'],category='General')

  #self.description				= options_item(None, 			"A description of the project." , 'entry','Description')	
  self.EGARCH	            = options_item(False,			'Normal GARCH, as opposed to EGARCH if True', bool, 'Estimate GARCH directly',
                                                      [True,False],['Direct GARCH','Usual GARCH'],category='ARIMA-GARCH')	


  self.do_not_constrain			= options_item(None, 			"The name of a variable of interest \nthat shall not be constrained due to \nmulticollinearity",
                                                                    [str,type(None)],"Avoid constraint",
                                                                                                                                         descr_for_input_boxes=['Variable not to constraint:'])	

  self.fixed_random_group_eff		= options_item(0,				'Fixed, random or no group effects', int, 'Group fixed random effect',[0,1,2], 
                                                                  ['No effects','Fixed effects','Random effects'],category='Fixed-random effects')
  self.fixed_random_time_eff		= options_item(0,				'Fixed, random or no time effects', int, 'Time fixed random effect',[0,1,2], 
                                                                 ['No effects','Fixed effects','Random effects'],category='Fixed-random effects')
  self.fixed_random_variance_eff	= options_item(0,				'Fixed, random or no group effects for variance', int, 'Variance fixed random effects',[0,1,2], 
                                                             ['No effects','Fixed effects','Random effects'],category='Fixed-random effects')



  self.h_function					= options_item(					"def h(e,z):\n"
                                                                                                                      "	e2			=	e**2+1e-5\n"
                                                                                                                                        "	h_val		=	np.log(e2)\n"	
                                                                                                                                        "	h_e_val		=	2*e/e2\n"
                                                                                                                                        "	h_2e_val	=	2/e2-4*e**2/e2**2\n"
                                                                                                                                        "	return h_val,h_e_val,h_2e_val,None,None,None\n",	

                                                                                                                                        "You can supply your own heteroskedasticity function. It must be a function of\n"
                                                                                                                                        "residuals e and a shift parameter z that is determined by the maximization procedure\n"
                                                                                                                                        "the function must return the value and its computation in the following order:\n"
                                                                                                                                        "h, dh/de, (d^2)h/de^2, dh/dz, (d^2)h/dz^2,(d^2)h/(dz*de)"
                                                                                                                                        , str,"GARCH function",category='Regression')
  self.include_initvar				= options_item(False,			"If True, includes an initaial variance term",
                                                                         bool,'Include initial variance', [True,False],['Include','Do not include'],category='Regression')

  self.initial_arima_garch_params = options_item(0.1,			'The initial size of arima-garch parameters (all directions will be attempted', 
                                                       float, 'initial size of arima-garch parameters',
                                                                                                                                                   "%s>=0",category='ARIMA-GARCH')		

  self.kurtosis_adj				= options_item(0,				'Amount of kurtosis adjustment', float, 'Amount of kurtosis adjustment',
                                                                        "%s>=0",category='ARIMA-GARCH')	

  self.GARCH_assist				= options_item(0,				'Amount of weight put on assisting GARCH variance to be close to squared residuals', float, 'GARCH assist',
                                                                        "%s>=0",category='ARIMA-GARCH')		

  self.min_group_df				= options_item(1, 				"The smallest permissible number of observations in each group. Must be at least 1", int, 'Minimum degrees of freedom', "%s>0",category='Regression')

  self.max_iterations			= options_item(10000, 			"Maximum number of iterations", int, 'Maximum number of iterations', "%s>0",category='Regression')
  
  self.max_increments			= options_item(0, 			"Maximum increment before maximization is ended", float, 'Maximum increments', "%s>0",category='Regression')

  self.minimum_iterations		= options_item(0, 				'Minimum number of iterations in maximization:',
                                                                      int,"Minimum iterations", "%s>-1")		


  self.pool						= options_item(False, 			"True if sample is to be pooled, otherwise False." 
                                                                                "For running a pooled regression",  
                                                                                                                                        bool,'Pooling',[True,False],['Pooled','Not Pooled'])

  self.pqdkm						= options_item([1,1,0,1,1], 
                                                                                 "ARIMA-GARCH parameters:",int, 'ARIMA-GARCH orders',
                                                                                                                                        ["%s>=0","%s>=0","%s>=0","%s>=0","%s>=0"],
                                                                                                                                        descr_for_input_boxes=["Auto Regression order (ARIMA, p)",
                                                                                                                                                               "Moving Average order (ARIMA, q)",
                                                                                                                                        "difference order (ARIMA, d)",
                                                                                                                                        "Variance Moving Average order (GARCH, k)",
                                                                                                                                        "Variance Auto Regression order (GARCH, m)"],category='Regression')

  self.robustcov_lags_statistics	= options_item([100,30],		"Numer of lags used in calculation of the robust \ncovariance matrix for the time dimension", 
                                                             int, 'Robust covariance lags (time)', ["%s>1","%s>1"], 
                                                                                                                                         descr_for_input_boxes=["# lags in final statistics calulation",
                                                                                                                                                                    "# lags iterations (smaller saves time)"],
                                                                                                                                             category='Output')

  self.silent						= options_item(False, 			"True if silent mode, otherwise False." 
                                                                                  "For running the procedure in a script, where output should be suppressed",  
                                                                                                                                        bool,'Silent mode',[True,False],['Silent','Not Silent'])

  self.subtract_means				= options_item(False,			"If True, subtracts the mean of all variables. This may be a remedy for multicollinearity if the mean is not of interest.",
                                                                          bool,'Subtract means', [True,False],['Subtracts the means','Do not subtract the means'],category='Regression')

  self.supress_output				= options_item(False,			"If True, no output is printed.",
                                                                          bool,'Supress output', [True,False],['Supresses output','Do not supress output'],category='Regression')

  self.tobit_limits				= options_item([None,None],		"Determines the limits in a tobit regression. "
                                                                        "Element 0 is lower limit and element1 is upper limit. "
                                                                                                                                        "If None, the limit is not active", 
                                                                                                                                        [float,type(None)], 'Tobit-model limits', 
                                                                                                                                        descr_for_input_boxes=['lower limit','upper limit'])

  self.tolerance					= options_item(0.000001, 		"Tolerance. When the maximum absolute value of the gradient divided by the hessian diagonal"
                                                                             "is smaller than the tolerance, the procedure is "
                                                                                                                                        "determined to have converged.",
                                                                                                                                        float,"Tolerance", "%s>0")	

  self.variance_RE_norm			= options_item(0.000005, 		"This parameter determines at which point the log function involved in the variance RE/FE calculations, "
                                                                    "will be extrapolate by a linear function for smaller values",
                                                                                                                                        float,"Variance RE/FE normalization point in log function", "%s>0")		
  self.user_constraints			= options_item(None,			"You can add constraints as a dict or as a string in python dictonary syntax.\n",
                                                                    [str,dict], 'User constraints')

  self.use_analytical		= options_item(1,				'Use analytical Hessian', int, 'Analytical Hessian',[0,1,2], 
                                                                  ['No analytical','Analytical in iterations only','Analytical initially and in iterations'],category='Genereal')

  self.make_category_tree()

  self.web_open_tab				= options_item(True, 			"True if web a new web browser tab should be opened when using web interface" 
                                                                        "Should a new tab be opemed?",  
                                                                                                                                        bool,'New web tab',[True,False],['Yes','No'])	

  return self


def application_preferences():
  opt=options()

  opt.save_datasets	= options_item(True, "If True, all loaded datasets are saved on exit and will reappear when the application is restarted", 
                                                bool,"Save datasets on exit", [False,True],
                                                                        ['Save on exit',
                                                                         'No thanks'])

  opt.n_round	= options_item(4, "Sets the number of digits the results are rounded to", 
                                          str,"Rounding digits", ['no rounding','0 digits','1 digits','2 digits','3 digits',
                                                                                                '4 digits','5 digits','6 digits','7 digits','8 digits',
                                                                                                                                                                                 '9 digits','10 digits'])

  opt.n_digits	= options_item(10, "Sets the maximum number of digits (for scientific format) if 'Rounding digits' is not set (-1)", 
                                           int,"Number of digits", ['0 digits','1 digits','2 digits','3 digits',
                                                                                                 '4 digits','5 digits','6 digits','7 digits','8 digits',
                                                                                                                                                                                 '9 digits','10 digits'])	
  opt.round_scientific	= options_item(True, "Determines if small numbers that are displayed in scientific format shall be rounded", 
                                                   bool,"Round Scientific", [True,False],['Round Scientific','Do not round scientific'])		
  opt.make_category_tree()

  return opt





#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .. import likelihood as logl
from . import direction
import numpy as np

STPMX=100.0 
import time
class LineSearch:
  def __init__(self, x, comput, panel, ll_old, step = 1):
    self.alf = 1.0e-3     #Ensures sufficient decrease in function value.
    self.tolx = 1.0e-14  #Convergence criterion on fx.		
    self.step = step
    self.stpmax = STPMX * max((abs(np.sum(x**2)	))**0.5,len(x))
    self.comput = comput
    self.panel = panel
    self.applied_constraints = []
    self.ll_old = ll_old

  def lnsrch(self, x, f, g, H, dx):

    #(x, f, g, dx) 

    self.conv=0
    self.rev = False
    self.g = g
    self.msg = ""
    n=len(x)
  
    if f is None:
      raise RuntimeError('f cannot be None')

    summ=np.sum(dx**2)**0.5
    if summ > self.stpmax:
      dx = dx*self.stpmax/summ 
    if np.all(dx==0):
      self.default(f, x, 0,  "dx is zero", 4)  
      return      
      
    test=0.0 															#Compute lambda min.
    for i in range(0,n): 
      temp=abs(dx[i])/max(abs(x[i]),1.0) 
      if (temp > test): test=temp 
    alamin = self.tolx/test 
    #*******CUSTOMIZATION
    #multithread:

    for i in range(1000):#Setting alam so that the largest step is valid. Set func to return None when input is invalid
      self.alam = 0.5**i*self.step #Always try full Newton step first.
      dx_alam, slope, self.rev, self.applied_constraints  = direction.new(g, x, H, self.comput.constr, f, dx, self.alam)
      self.x = x + dx_alam
      self.f, self.ll = self.func(self.x) 
      if self.f != None: 
        break
    #*************************
    f2=0
    alam2 = self.alam
    alamstart = self.alam#***********CUSTOMIZATION
    max_iter = 1000
    for self.k in range (0,max_iter):			#Start of iteration loop.
      dx_alam, slope, self.rev, self.applied_constraints  = direction.new(g, x, H, self.comput.constr, f, dx, self.alam)
      self.x = x + dx_alam
      if self.k > 0: 
        self.f, self.ll = self.func(self.x) 
      if self.f is None:
        self.msg = 'The function returned None'
        self.f = f
      if (self.alam < alamin):   #Convergence on delta x. For zero finding,the calling program should verify the convergence.
        self.default(f, x, 0, "Convergence on delta dx", 2)
        return
      elif (self.f >= f+self.alf*self.alam*slope): 
        self.msg = "Sufficient function increase"
        #print(f'LL single:{self.ll.LL}')
        #self.ll = self.ll2
        self.conv = 1
        return							#Sufficient function increase
      else:  															#Backtrack.
        if (self.alam == alamstart):#***********CUSTOMIZATION  alam == 1.0
          tmplam = -slope/(2.0*(self.f-f-slope))  	#First time.
        else:  														#Subsequent backtracks.
          rhs1 = self.f-f-self.alam*slope 
          rhs2 = f2-f-alam2*slope 
          a=(rhs1/(self.alam**2)-rhs2/(alam2*alam2))/(self.alam-alam2) 
          b=(-alam2*rhs1/(self.alam**2)+self.alam*rhs2/(alam2*alam2))/(self.alam-alam2) 
          if (a == 0.0):
            tmplam = -slope/(2.0*b)  
          else:  
            disc=b*b-3.0*a*slope 
            if (disc < 0.0):
              tmplam = 0.5*self.alam  
            elif (b >= 0.0):
              tmplam=-(b+(disc)**0.5)/(3.0*a) 
            else:
              tmplam=slope/(-b+(disc)**0.5)
          if (tmplam > 0.5*self.alam): 
            tmplam = 0.5*self.alam   								#  lambda<=0.5*lambda1
      alam2 = self.alam 
      f2 = self.f
      self.alam = max(tmplam, 0.1*self.alam)								#lambda>=0.1*lambda1
      if alamstart<1.0:#*************CUSTOMIZATION
        self.alam = min((self.alam, alamstart*0.9**self.k))
      
    self.msg = f"No function increase after {max_iter} iterations"
    self.conv = 3

  def func(self,x):	
    ll = logl.LL(x, self.panel, self.comput.constr)
    if ll is None:
      return None, None
    elif ll.LL is None:
      return None, None
    return ll.LL, ll

  def default(self, f, x, alam, msg, conv):
    self.msg = msg
    self.conv = conv
    self.f = f
    self.x = x
    self.alam = alam
    self.ll = self.ll_old



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from .. import functions as fu
from . import calculus_functions as cf
from . import function
import numpy as np
import time

class gradient:

  def __init__(self,panel,callback):
    self.panel=panel
    self.callback=callback

  def arima_grad(self,k,x,ll,sign,pre):
    if k==0:
      return None
    (N,T,m)=x.shape
    x=self.panel.arma_dot.dotroll(pre,k,sign,x,ll)
    x.resize(N,T,k)
    extr_value=1e+100
    if np.max(np.abs(x))>extr_value:
      x[np.abs(x)>extr_value]=np.sign(x[np.abs(x)>extr_value])*extr_value
    return x*self.panel.a[3]

  def garch_arima_grad(self,ll,dRE,varname):
    panel=self.panel
    groupeffect=0
    groupeffect, dvRE_dx=None, None
    d_input=0
    if self.panel.N>1 and panel.options.fixed_random_group_eff.value>0 and not dRE is None:
      d_eRE_sq=2*ll.e_RE*dRE
      dmeane2=panel.mean(d_eRE_sq,(0,1))
      d_input=(d_eRE_sq-dmeane2)*panel.a[3]
      dvRE_dx=dmeane2*panel.a[3]-ll.re_obj_i_v.dRE(d_input,ll.varRE_input,varname,panel)-ll.re_obj_t_v.dRE(d_input,ll.varRE_input,varname,panel)
      groupeffect=ll.dvarRE*dvRE_dx*panel.a[3]


    dvar_sigma=None
    if self.panel.pqdkm[4]>0 and not dRE is None: 			#eqs. 33-34
      ((N,T,k))=dRE.shape
      x=cf.prod((ll.h_e_val,dRE))	
      dvar_sigma=self.panel.arma_dot.dot(ll.GAR_1MA,x,ll)
    dvar_e=cf.add((dvar_sigma,groupeffect),True)
    return dvar_e,dvar_sigma,dvRE_dx,d_input


  def get(self,ll,DLL_e=None,dLL_var=None,return_G=False):


    self.callback(perc = 0.05, text = '', task = 'gradient')
    (self.DLL_e, self.dLL_var)=(DLL_e, dLL_var)
    panel=self.panel
    incl=self.panel.included[3]
    re_obj_i,re_obj_t=ll.re_obj_i,ll.re_obj_t
    u, e_RE,u_RE,h_e_val,var,h_val,v=ll.u, ll.e_RE,ll.u_RE,ll.h_e_val,ll.var,ll.h_val,ll.v
    p,q,d,k,m=panel.pqdkm
    nW=panel.nW
    if DLL_e is None:
      dLL_var, DLL_e=function.gradient(ll,self.panel)
    self.X_RE = (panel.XIV+re_obj_i.RE(panel.XIV, panel)+re_obj_t.RE(panel.XIV, panel))*panel.included[3]
    #ARIMA:
    de_rho_RE=self.arima_grad(p,u_RE,ll,-1,ll.AMA_1)
    de_lambda_RE=self.arima_grad(q,e_RE,ll,-1,ll.AMA_1)
    de_beta_RE=-self.panel.arma_dot.dot(ll.AMA_1AR,self.X_RE,ll)*panel.a[3]

    (self.de_rho_RE,self.de_lambda_RE,self.de_beta_RE)=(de_rho_RE,de_lambda_RE,de_beta_RE)		

    dvar_sigma_rho,		dvar_sigma_rho_G,		dvRE_rho	, d_rho_input		=	self.garch_arima_grad(ll,	self.de_rho_RE,		'rho')
    dvar_sigma_lambda, 	dvar_sigma_lambda_G,	dvRE_lambda	, d_lambda_input	=	self.garch_arima_grad(ll,	self.de_lambda_RE,	'lambda')
    dvar_sigma_beta,	dvar_sigma_beta_G,		dvRE_beta	, d_beta_input		=	self.garch_arima_grad(ll,	self.de_beta_RE,	'beta')


    (self.dvar_sigma_rho,self.dvar_sigma_lambda,self.dvar_sigma_beta)=(dvar_sigma_rho,dvar_sigma_lambda,dvar_sigma_beta)
    (self.dvar_sigma_rho_G,self.dvar_sigma_lambda_G,self.dvar_sigma_beta_G)=(dvar_sigma_rho_G,dvar_sigma_lambda_G,dvar_sigma_beta_G)
    (self.dvRE_rho,self.dvRE_lambda,self.dvRE_beta)=(dvRE_rho,dvRE_lambda,dvRE_beta)
    (self.d_rho_input,self.d_lambda_input,self.d_beta_input)=(d_rho_input,d_lambda_input,d_beta_input)

    #GARCH:

    self.dvar_omega=self.panel.arma_dot.dot(ll.GAR_1,panel.W_a,ll)
    self.dvar_initvar = None
    if 'initvar' in ll.args.args_d:
      self.dvar_initvar = ll.GAR_1[0].reshape((1,panel.max_T,1))
      if not panel.options.EGARCH.value and ll.args.args_d['initvar'][0]<0:	
        self.dvar_initvar = -self.dvar_initvar
      
    (dvar_gamma, dvar_psi, dvar_mu, dvar_z_G, dvar_z)=(None,None,None,None,None)
    
    if panel.N>1:
      dvar_mu=cf.prod((ll.dvarRE_mu,incl))
    else:
      dvar_mu=None	

    if m>0:
      dvar_gamma=self.arima_grad(k,var,ll,1,ll.GAR_1)
      dvar_psi=self.arima_grad(m,h_val,ll,1,ll.GAR_1)
      if not ll.h_z_val is None:
        dvar_z_G=fu.dot(ll.GAR_1MA,ll.h_z_val)
        (N,T,k)=dvar_z_G.shape

      dvar_z=dvar_z_G


    (self.dvar_gamma, self.dvar_psi,self.dvar_mu,self.dvar_z_G,self.dvar_z)=(dvar_gamma, dvar_psi, dvar_mu, dvar_z_G, dvar_z)

    #LL


    #final derivatives:
    dLL_beta=cf.add((cf.prod((dvar_sigma_beta,dLL_var)),cf.prod((self.de_beta_RE,DLL_e))),True)
    dLL_rho=cf.add((cf.prod((dvar_sigma_rho,dLL_var)),cf.prod((self.de_rho_RE,DLL_e))),True)
    dLL_lambda=cf.add((cf.prod((dvar_sigma_lambda,dLL_var)),cf.prod((self.de_lambda_RE,DLL_e))),True)
    dLL_gamma=cf.prod((dvar_gamma,dLL_var))
    dLL_psi=cf.prod((dvar_psi,dLL_var))
    dLL_omega=cf.prod((self.dvar_omega,dLL_var))
    dLL_initvar=cf.prod((self.dvar_initvar,dLL_var))
    dLL_mu=cf.prod((self.dvar_mu,dLL_var))
    dLL_z=cf.prod((self.dvar_z,dLL_var))


    G=cf.concat_marray((dLL_beta,dLL_rho,dLL_lambda,dLL_gamma,dLL_psi,dLL_omega, dLL_initvar,dLL_mu,dLL_z))
    g=np.sum(G,(0,1))
    #For debugging:
    #from .. import debug
    #print(debug.grad_debug(ll,panel,0.00001))
    #print(g)
    #if np.sum((g-gn)**2)>10000000:
    #	a=0
    #print(gn)
    #a=debug.grad_debug_detail(ll, panel, 0.00000001, 'LL', 'beta',0)
    #dLLeREn,deREn=debug.LL_calc_custom(ll, panel, 0.0000001)

    self.callback(perc = 0.08, text = '', task = 'gradient')
    


    if return_G:
      return  g,G
    else:	
      return g

class hessian:
  def __init__(self,panel,g,callback):
    self.panel=panel
    self.its=0
    self.g=g
    self.callback=callback


  def get(self,ll,d2LL_de2,d2LL_dln_de,d2LL_dln2):	
    H = self.hessian(ll,d2LL_de2,d2LL_dln_de,d2LL_dln2)
    return H


  def hessian(self,ll,d2LL_de2,d2LL_dln_de,d2LL_dln2):
    panel=self.panel
    tic=time.perf_counter()
    g=self.g
    p,q,d,k,m=panel.pqdkm
    incl=self.panel.included[3]

    GARM=(ll.GAR_1,m,1)

    GARK=(ll.GAR_1,k,1)

    d2var_gamma2					=   cf.prod((2, 
                                       cf.dd_func_lags(panel,ll,GARK, 	g.dvar_gamma,						g.dLL_var,  transpose=True)))
    d2var_gamma_psi				=	cf.dd_func_lags(panel,ll,GARK, 	g.dvar_psi,							g.dLL_var)
    d2var_gamma_rho				=	cf.dd_func_lags(panel,ll,GARK,	g.dvar_sigma_rho_G,			g.dLL_var)
    d2var_gamma_lambda		=	cf.dd_func_lags(panel,ll,GARK, 	g.dvar_sigma_lambda_G,	g.dLL_var)
    d2var_gamma_beta			=	cf.dd_func_lags(panel,ll,GARK, 	g.dvar_sigma_beta_G,		g.dLL_var)
    d2var_gamma_initvar 	= cf.dd_func_lags(panel,ll,GARK, 	g.dvar_initvar,					g.dLL_var)
    d2var_gamma_omega 		= cf.dd_func_lags(panel,ll,GARK, 	g.dvar_omega,					g.dLL_var)
    d2var_gamma_z					=	cf.dd_func_lags(panel,ll,GARK, 	g.dvar_z_G,							g.dLL_var)
    
    self.callback(perc = 0.2, text = '', task = 'hessian')
    
    d2var_psi_rho					=	cf.dd_func_lags(panel,ll,GARM, 	cf.prod((ll.h_e_val,g.de_rho_RE)),		g.dLL_var)
    d2var_psi_lambda			=	cf.dd_func_lags(panel,ll,GARM, 	cf.prod((ll.h_e_val,g.de_lambda_RE)),	g.dLL_var)
    d2var_psi_beta				=	cf.dd_func_lags(panel,ll,GARM, 	cf.prod((ll.h_e_val,g.de_beta_RE)),	g.dLL_var)
    d2var_psi_z						=	cf.dd_func_lags(panel,ll,GARM, 	ll.h_z_val,								g.dLL_var)

    AMAq=(ll.AMA_1,q,-1)
    d2var_lambda2,		d2e_lambda2		=	cf.dd_func_lags_mult(panel,ll,g,AMAq,	'lambda_RE',	'lambda_RE', transpose=True)
    d2var_lambda_rho,	d2e_lambda_rho	=	cf.dd_func_lags_mult(panel,ll,g,AMAq,	'lambda_RE',	'rho_RE' )
    d2var_lambda_beta,	d2e_lambda_beta	=	cf.dd_func_lags_mult(panel,ll,g,AMAq,	'lambda_RE',	'beta_RE')

    AMAp=(ll.AMA_1,p,-1)
    d2var_rho_beta,		d2e_rho_beta	=	cf.dd_func_lags_mult(panel,ll,g,AMAp,	'rho_RE',		'beta_RE', u_gradient=True)

    self.callback(perc = 0.4, text = '', task = 'hessian')

    d2var_mu_rho,d2var_mu_lambda,d2var_mu_beta,d2var_mu_z,mu=None,None,None,None,None
    if panel.N>1:
      d2var_mu_rho			=	cf.sumNT(cf.prod((ll.ddvarRE_mu_vRE, 	g.dvRE_rho,  	 	g.dLL_var)))
      d2var_mu_lambda			=	cf.sumNT(cf.prod((ll.ddvarRE_mu_vRE, 	g.dvRE_lambda,  	g.dLL_var)))
      d2var_mu_beta			=	cf.sumNT(cf.prod((ll.ddvarRE_mu_vRE, 	g.dvRE_beta,  	 	g.dLL_var)))
      d2var_mu_z=None
      d2var_mu2=0

    self.callback(perc = 0.5, text = '', task = 'hessian')
    d2var_z2				=	cf.dd_func_lags(panel,ll,ll.GAR_1MA, ll.h_2z_val,						g.dLL_var) 
    d2var_z_rho				=	cf.dd_func_lags(panel,ll,ll.GAR_1MA, cf.prod((ll.h_ez_val,g.de_rho_RE)),	g.dLL_var) 
    d2var_z_lambda			=	cf.dd_func_lags(panel,ll,ll.GAR_1MA, cf.prod((ll.h_ez_val,g.de_lambda_RE)),g.dLL_var) 
    d2var_z_beta			=	cf.dd_func_lags(panel,ll,ll.GAR_1MA, cf.prod((ll.h_ez_val,g.de_beta_RE)),	g.dLL_var) 

    d2var_rho2,	d2e_rho2	=	cf.dd_func_lags_mult(panel,ll,g,	None,	'rho_RE',		'rho_RE' )
    d2var_beta2,d2e_beta2	=	cf.dd_func_lags_mult(panel,ll,g,	None,	'beta_RE',		'beta_RE')



    (de_rho_RE,de_lambda_RE,de_beta_RE)=(g.de_rho_RE,g.de_lambda_RE,g.de_beta_RE)
    (dvar_sigma_rho,dvar_sigma_lambda,dvar_sigma_beta)=(g.dvar_sigma_rho,g.dvar_sigma_lambda,g.dvar_sigma_beta)
    (dvar_mu,dvar_z)=(g.dvar_mu, g.dvar_z)		

    d2var_beta_omega, d2var_beta_initvar, d2var_rho_omega, d2var_lambda_omega=None, None, None, None
    d2var_rho_initvar, d2var_lambda_initvar = None, None
    

    self.callback(perc = 0.6, text = '', task = 'hessian')
    #Final:
    D2LL_beta2 						=	cf.dd_func(d2LL_de2,	d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	de_beta_RE,		dvar_sigma_beta, 	dvar_sigma_beta,	d2e_beta2, 					d2var_beta2)
    D2LL_beta_rho		      =	cf.dd_func(d2LL_de2,	d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	de_rho_RE,		dvar_sigma_beta, 	dvar_sigma_rho,		T(d2e_rho_beta), 		T(d2var_rho_beta))
    D2LL_beta_lambda			=	cf.dd_func(d2LL_de2,	d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	de_lambda_RE,	dvar_sigma_beta, 	dvar_sigma_lambda,	T(d2e_lambda_beta), 	T(d2var_lambda_beta))
    D2LL_beta_gamma				=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	None,			dvar_sigma_beta, 	g.dvar_gamma,		None, 					T(d2var_gamma_beta))
    D2LL_beta_psi					=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	None,			dvar_sigma_beta, 	g.dvar_psi,			None, 					T(d2var_psi_beta))
    D2LL_beta_omega				=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	None,			dvar_sigma_beta, 	g.dvar_omega,		None, 					d2var_beta_omega)
    D2LL_beta_initvar			=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	None,			dvar_sigma_beta, 	g.dvar_initvar,	None, 					d2var_beta_initvar)
    D2LL_beta_mu					=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	None,			dvar_sigma_beta, 	dvar_mu,				None, 					d2var_mu_beta)
    D2LL_beta_z						=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	None,			dvar_sigma_beta, 	dvar_z,					None, 					T(d2var_z_beta))

    D2LL_rho2							=	cf.dd_func(d2LL_de2,	d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		de_rho_RE,		dvar_sigma_rho, 	dvar_sigma_rho,		d2e_rho2, 					d2var_rho2)
    D2LL_rho_lambda				=	cf.dd_func(d2LL_de2,	d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		de_lambda_RE,	dvar_sigma_rho, 	dvar_sigma_lambda,	T(d2e_lambda_rho), 		T(d2var_lambda_rho))
    D2LL_rho_gamma				=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		None,			dvar_sigma_rho, 	g.dvar_gamma,		None, 					T(d2var_gamma_rho))	
    D2LL_rho_psi					=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		None,			dvar_sigma_rho, 	g.dvar_psi,			None, 					T(d2var_psi_rho))
    D2LL_rho_omega				=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		None,			dvar_sigma_rho, 	g.dvar_omega,		None, 					d2var_rho_omega)
    D2LL_rho_initvar			=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 	 	None,			dvar_sigma_rho, 	g.dvar_initvar,	None, 					d2var_rho_initvar)
    D2LL_rho_mu						=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		None,			dvar_sigma_rho, 	dvar_mu,				None, 					T(d2var_mu_rho))
    D2LL_rho_z						=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		None,			dvar_sigma_rho, 	dvar_z,					None, 					T(d2var_z_rho))

    D2LL_lambda2					=	cf.dd_func(d2LL_de2,	d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	de_lambda_RE,	dvar_sigma_lambda, 	dvar_sigma_lambda,	T(d2e_lambda2), 		T(d2var_lambda2))
    D2LL_lambda_gamma			=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	None,			dvar_sigma_lambda, 	g.dvar_gamma,		None, 					T(d2var_gamma_lambda))
    D2LL_lambda_psi				=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	None,			dvar_sigma_lambda, 	g.dvar_psi,			None, 					T(d2var_psi_lambda))
    D2LL_lambda_omega			=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	None,			dvar_sigma_lambda, 	g.dvar_omega,		None, 					d2var_lambda_omega)
    D2LL_lambda_initvar		=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	None,			dvar_sigma_lambda, 	g.dvar_initvar,	None, 					d2var_lambda_initvar)
    D2LL_lambda_mu				=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	None,			dvar_sigma_lambda, 	dvar_mu,				None, 					T(d2var_mu_lambda))
    D2LL_lambda_z					=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	None,			dvar_sigma_lambda, 	dvar_z,					None, 					T(d2var_z_lambda))

    D2LL_gamma2						=	cf.dd_func(None,		None,			d2LL_dln2,	None, 		None,			g.dvar_gamma, 		g.dvar_gamma,		None, 					T(d2var_gamma2))
    D2LL_gamma_psi				=	cf.dd_func(None,		None,			d2LL_dln2,	None,			None,			g.dvar_gamma, 		g.dvar_psi,			None, 					d2var_gamma_psi)
    D2LL_gamma_omega			=	cf.dd_func(None,		None,			d2LL_dln2,	None, 		None,			g.dvar_gamma, 		g.dvar_omega,		None, 					d2var_gamma_omega)
    D2LL_gamma_initvar		=	cf.dd_func(None,		None,			d2LL_dln2,	None, 		None,			g.dvar_gamma,		 	g.dvar_initvar,	None, 					d2var_gamma_initvar)
    D2LL_gamma_mu					=	cf.dd_func(None,		None,			d2LL_dln2,	None, 		None,			g.dvar_gamma, 		dvar_mu,				None, 					None)
    D2LL_gamma_z					=	cf.dd_func(None,		None,			d2LL_dln2,	None, 		None,			g.dvar_gamma, 		dvar_z,					None, 					d2var_gamma_z)

    D2LL_psi2							=	cf.dd_func(None,		None,			d2LL_dln2,	None,			None,			g.dvar_psi, 		g.dvar_psi,			None, 					None)
    D2LL_psi_omega				=	cf.dd_func(None,		None,			d2LL_dln2,	None, 		None,			g.dvar_psi, 		g.dvar_omega,		None, 					None)
    D2LL_psi_initvar			=	cf.dd_func(None,		None,			d2LL_dln2,	None, 		None,			g.dvar_psi, 	 	g.dvar_initvar,	None, 					None)
    D2LL_psi_mu						=	cf.dd_func(None,		None,			d2LL_dln2,	None, 		None,			g.dvar_psi, 		dvar_mu,			 	None, 					None)
    D2LL_psi_z						=	cf.dd_func(None,		None,			d2LL_dln2,	None, 		None,			g.dvar_psi, 		dvar_z,				 	None, 					d2var_psi_z)

    D2LL_omega2						=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_omega, 	g.dvar_omega,		None, 					None)
    D2LL_omega_initvar		=	cf.dd_func(None,		None,	 	 	d2LL_dln2,	None, 	 	  None,			g.dvar_omega, 	g.dvar_initvar,	None, 					None)
    D2LL_omega_mu					=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_omega, 	g.dvar_mu,			None, 					None)
    D2LL_omega_z					=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_omega, 	g.dvar_z,			 	None, 					None)

    D2LL_initvar2					=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_initvar,		g.dvar_initvar,		None, 					None)
    D2LL_initvar_mu				=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			dvar_mu, 					g.dvar_initvar,		None, 					None)
    D2LL_initvar_z				=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			dvar_z, 					g.dvar_initvar,		None, 					None)

    D2LL_mu2							=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			dvar_mu, 			dvar_mu,			None, 					None)
    D2LL_mu_z							=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			dvar_mu, 			dvar_z,				None, 					d2var_mu_z)

    D2LL_z2								=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			dvar_z, 			dvar_z,				None, 					d2var_z2)




    H= [[D2LL_beta2,					D2LL_beta_rho,				D2LL_beta_lambda,				D2LL_beta_gamma,			D2LL_beta_psi,			D2LL_beta_omega,				D2LL_beta_initvar, 		D2LL_beta_mu,		D2LL_beta_z		],
        [T(D2LL_beta_rho),		D2LL_rho2,						D2LL_rho_lambda,				D2LL_rho_gamma,				D2LL_rho_psi,				D2LL_rho_omega,					D2LL_rho_initvar, 		D2LL_rho_mu,		D2LL_rho_z			],
        [T(D2LL_beta_lambda),	T(D2LL_rho_lambda),		D2LL_lambda2,						D2LL_lambda_gamma,		D2LL_lambda_psi,		D2LL_lambda_omega,			D2LL_lambda_initvar, 	D2LL_lambda_mu,	D2LL_lambda_z		],
        [T(D2LL_beta_gamma),	T(D2LL_rho_gamma),		T(D2LL_lambda_gamma),		D2LL_gamma2,					D2LL_gamma_psi,			D2LL_gamma_omega, 			D2LL_gamma_initvar, 	D2LL_gamma_mu,	D2LL_gamma_z		],
        [T(D2LL_beta_psi),		T(D2LL_rho_psi),			T(D2LL_lambda_psi),			T(D2LL_gamma_psi),		D2LL_psi2,					D2LL_psi_omega, 				D2LL_psi_initvar, 		D2LL_psi_mu,		D2LL_psi_z			],
        [T(D2LL_beta_omega),	T(D2LL_rho_omega),		T(D2LL_lambda_omega),		T(D2LL_gamma_omega),	T(D2LL_psi_omega),	D2LL_omega2, 						D2LL_omega_initvar, 	D2LL_omega_mu,	D2LL_omega_z		], 
        [T(D2LL_beta_initvar),T(D2LL_rho_initvar),	T(D2LL_lambda_initvar),	T(D2LL_gamma_initvar),T(D2LL_psi_initvar),T(D2LL_omega_initvar), 	D2LL_initvar2, 				D2LL_initvar_mu,D2LL_initvar_z		], 
        [T(D2LL_beta_mu),			T(D2LL_rho_mu),				T(D2LL_lambda_mu),			T(D2LL_gamma_mu),			T(D2LL_psi_mu),			T(D2LL_omega_mu), 			D2LL_initvar_mu, 			D2LL_mu2,				D2LL_mu_z			],
        [T(D2LL_beta_z),			T(D2LL_rho_z),				T(D2LL_lambda_z),				T(D2LL_gamma_z),			T(D2LL_psi_z),			T(D2LL_omega_z), 				D2LL_initvar_z, 			D2LL_mu_z,			D2LL_z2				]]
    self.callback(perc = 0.8, text = '', task = 'hessian')
    H=cf.concat_matrix(H)
    if H[-1,-1]==0:
      H[-1,-1]=1
    #for debugging:
    #from .. import debug
    #Hn=debug.hess_debug(ll,panel,g,0.00000001)#debugging
    #v=debug.hess_debug_detail(ll,panel,0.0000001,'grp','beta','beta',0,0)
    #print (time.perf_counter()-tic)
    self.its+=1
    if np.any(np.isnan(H)):
      return None
    #print(H[0]/1e+11)
    return H





def T(x):
  if x is None:
    return None
  return x.T


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pydoc import importfile
import os
path = os.path.dirname(__file__)
from ..output import stat_functions as stat


import numpy as np



class Constraint:
  def __init__(self,index,assco,cause,value, interval,names,category):
    self.name=names[index]
    self.intervalbound=None
    self.max=None
    self.min=None
    self.value=None
    self.value_str=None
    if interval is None:
      self.value=value
      self.value_str=str(round(self.value,8))
    else:
      if interval[0]>interval[1]:
        raise RuntimeError('Lower constraint cannot exceed upper')
      self.min=interval[0]
      self.max=interval[1]
      self.cause='user/general constraint'
    self.assco_ix=assco
    if assco is None:
      self.assco_name=None
    else:
      self.assco_name=names[assco]
    self.cause=cause
    self.category=category	

class Constraints(dict):

  """Stores the constraints of the LL maximization"""	
  def __init__(self,panel,args,its=0):
    dict.__init__(self)
    self.categories={}
    self.fixed={}
    self.intervals={}
    self.associates={}
    self.collinears={}
    self.weak_mc_dict={}
    self.args=args
    self.args[0]
    self.panel_args=panel.args
    self.CI=None
    self.its=its
    self.pqdkm=panel.pqdkm
    self.m_zero=panel.m_zero
    self.ARMA_constraint=panel.options.ARMA_constraint.value
    self.H_correl_problem=False
    self.mc_problems=[]	

  def add(self,name,assco,cause,interval=None,replace=True,value=None,args=None):
    #(self,index,assco,cause,interval=None,replace=True,value=None)
    name,index=self.panel_args.get_name_ix(name)
    name_assco,assco=self.panel_args.get_name_ix(assco,True)
    for i in index:
      self.add_item(i,assco,cause, interval ,replace,value,args)

  def clear(self,cause=None):
    for c in list(self.keys()):
      if self[c].cause==cause or cause is None:
        self.delete(c)	

  def add_item(self,index,assco,cause,interval,replace,value,args):
    """Adds a constraint. 'index' is the position
    for which the constraints shall apply.  \n\n

    Equality constraints are chosen by specifying 'minimum_or_value' \n\n
    Inequality constraints are chosen specifiying 'maximum' and 'minimum'\n\n
    'replace' determines whether an existing constraint shall be replaced or not 
    (only one equality and inequality allowed per position)"""

    if args is None:
      args=self.panel_args

    if not replace:
      if index in self:
        return False
    interval,value=test_interval(interval, value)
    if interval is None:#this is a fixed constraint
      if len(self.fixed)==len(args.caption_v)-1:#can't lock all variables
        return False
      if value is None:
        value=self.args[index]
      if index in self.intervals:
        c=self[index]
        if not (c.min<value<c.max):
          return False
        else:
          self.intervals.pop(index)
    elif index in self.fixed: #this is an interval constraint, no longer a fixed constraint
      self.fixed.pop(index)

    eq,category,j=self.panel_args.positions_map[index]
    if not category in self.categories:
      self.categories[category]=[index]
    elif not index in self.categories[category]:
      self.categories[category].append(index)

    c=Constraint(index,assco,cause,value, interval ,args.caption_v,category)
    self[index]=c
    if value is None:
      self.intervals[index]=c
    else:
      self.fixed[index]=c
    if not assco is None:
      if not assco in self.associates:
        self.associates[assco]=[index]
      elif not index in self.associates[assco]:
        self.associates[assco].append(index)
    if cause=='collinear':
      self.collinears[index]=assco
    return True

  def delete(self,index):
    if not index in self:
      return False
    self.pop(index)
    if index in self.intervals:
      self.intervals.pop(index)
    if index in self.fixed:
      self.fixed.pop(index)		
    eq,category,j=self.panel_args.positions_map[index]
    c=self.categories[category]
    if len(c)==1:
      self.categories.pop(category)
    else:
      i=np.nonzero(np.array(c)==index)[0][0]
      c.pop(i)
    a=self.associates
    for i in a:
      if index in a[i]:
        if len(a[i])==1:
          a.pop(i)
          break
        else:
          j=np.nonzero(np.array(a[i])==index)[0][0]
          a[i].pop(j)
    if index in self.collinears:
      self.collinears.pop(index)
    return True


  def set_fixed(self,x):
    """Sets all elements of x that has fixed constraints to the constraint value"""
    for i in self.fixed:
      x[i]=self.fixed[i].value

  def within(self,x,fix=False):
    """Checks if x is within interval constraints. If fix=True, then elements of
    x outside constraints are set to the nearest constraint. if fix=False, the function 
    returns False if x is within constraints and True otherwise"""
    for i in self.intervals:
      c=self.intervals[i]
      if (c.min<=x[i]<=c.max):
        c.intervalbound=None
      else:
        if fix:
          x[i]=max((min((x[i],c.max)),c.min))
          c.intervalbound=str(round(x[i],8))
        else:
          return False
    return True

  def add_static_constraints(self, panel, its, init_arma_constr, ll=None, minvarhits = []):
    pargs=self.panel_args
    p, q, d, k, m=self.pqdkm

    if its<-4:
      c = 0.5
    else:
      c=self.ARMA_constraint


    general_constraints=[('rho',-c,c),('lambda',-c,c),('gamma',-c,c),('psi',-c,c)]
    self.add_custom_constraints(panel, general_constraints, ll)
    self.add_custom_constraints(panel, pargs.user_constraints, ll)
    if len(minvarhits):
      self.add('gamma',None, 'Variance under threshold for an observation')
      self.add('psi',None, 'Variance under threshold for an observation')
      if not ll is None:
        for i in range(1,len(ll.args.args_d['omega'])):
          self.add(f'omega{i}',None, 'Variance under threshold for an observation')
          
    c = [
                  [(f'rho{i}', 0) for i in range(1,p)] +
                        [(f'lambda{i}', None) for i in range(1,q)] + 
                        [(f'gamma{i}', None) for i in range(1,k)] +
                        [(f'psi{i}', None) for i in range(1,m)],

                        [(f'rho{i}', None) for i in range(0,p)] +
                        [(f'lambda{i}', None) for i in range(0,q)] + 
                        [(f'gamma{i}', None) for i in range(0,k)] +
                        [(f'psi{i}', None) for i in range(0,m)]			
                ]
    if init_arma_constr>0:
      self.add_custom_constraints(panel, c[init_arma_constr-1], ll)
    a=0

  def add_dynamic_constraints(self,computation, H, ll, args = None):
    if not args is None:
      self.args = args
      self.args[0]
    k,k=H.shape
    self.weak_mc_dict=dict()
    include=np.array(k*[True])
    include[list(computation.constr.fixed)]=False
    self.mc_problems=[]#list of [index,associate,condition index]
    self.CI=constraint_multicoll(k, computation, include, self.mc_problems, H)
    self.add_mc_constraint(computation)

  def add_custom_constraints(self,panel, constraints, ll,replace=True,cause='user constraint',clear=False,args=None):
    """Adds custom range constraints\n\n
    	constraints shall be on the format [(name, minimum, maximum), ...]
    	or
    	a dictionary of the same form of an argument dictionary
    	(a dictionary of dictonaries on the form dict[category][name]) 
    	where items are lists of [minimum,maximum] or the fixing value

    	name is taken from self.panel_args.caption_v"""	

    if clear:
      self.clear(cause)
    if type(constraints)==list or type(constraints)==tuple:
      for c in constraints:
        self.add_custom_constraints_list(c,replace,cause,args, ll)
    elif type(constraints)==dict:
      self.add_custom_constraints_dict(panel, constraints,replace,cause,args)
    else:
      raise TypeError("The constraints must be a list, tuple or dict.")


  def add_custom_constraints_list(self,constraint,replace,cause,args, ll):
    """Adds a custom range constraint\n\n
    	constraint shall be on the format (name, minimum, maximum)"""
    if type(constraint)==str or isinstance(constraint,int):
      name, value=constraint,None
      self.add(name,None,cause,replace=replace,value=value,args=args)
      return
    elif len(constraint)==2:
      name, minimum,maximum=list(constraint)+[None]
    elif len(constraint)==3:
      name, minimum, maximum=constraint
    else:
      raise TypeError("A constraint needs to be a string, integer or iterable with lenght no more than tree.")

    name,ix=self.panel_args.get_name_ix(name)
    m=[minimum,maximum]
    for i in range(2):
      if type(m[i])==str:
        try:
          m[i]=eval(m[i],globals(),ll.__dict__)
        except:
          print(f"Custom constraint {name} ({m[i]}) failed")
          return
    [minimum,maximum]=m
    self.add(name,None,cause, [minimum,maximum],replace,args=args)

  def add_custom_constraints_dict(self, panel, constraints,replace,cause,args):
    """Adds a custom range constraint\n\n
       If list, constraint shall be on the format (minimum, maximum)"""
    for grp in constraints:
      for i in range(len(panel.args.caption_d[grp])):
        name = panel.args.caption_d[grp][i]
        c=constraints[grp]
        try:
          iter(c)
          self.add(name,None,cause, c,replace,args=args)
        except TypeError as e:
          self.add(name,None,cause, [c,None],replace,args=args)

  def add_mc_constraint(self, computation):
    """Adds constraints for severe MC problems"""
    old = computation.constr_old
    if len(self.mc_problems)==0:
      return

    no_check=get_no_check(computation)
    a=[i[0] for i in self.mc_problems]
    if no_check in a:
      mc=self.mc_problems[a.index(no_check)]
      mc[0],mc[1]=mc[1],mc[0]
    mc_limit = computation.panel.options.multicoll_threshold_report.value
    coll_max_limit = computation.panel.options.multicoll_threshold_max.value
    for index,assc,cond_index in self.mc_problems:
      cond = not ((index in self.associates) or (index in self.collinears))and (not index==no_check)
      # same condition, but possible have a laxer condition for weak_mc_dict. 
      if cond and cond_index>mc_limit :#contains also collinear variables that are only slightly collinear, which shall be restricted when calcuating CV-matrix:	
        self.weak_mc_dict[index]=[assc,cond_index]
      if cond and cond_index > coll_max_limit :#adding restrictions:
        self.add(assc,index,'collinear')
        return
        if old is None:
          self.add(assc,index,'collinear')
        elif assc in old.collinears:
          self.add(index ,assc,'collinear')
        else:
          self.add(assc,index,'collinear')
          
  def print_constraints(self, kind = None):
    for desc, obj in [('All', self),
                ('Fixed', self.fixed),
                ('Intervals', self.intervals)]:
      print(f"{desc} constraints:")
      for i in obj:
        c=obj[i]
        try:
          print(f"constraint: {i}, associate:{c.assco_ix}, max:{c.max}, min:{c.min}, value:{c.value}")  
        except:
          print(f"constraint: {i}, associate:{c.assco_ix}, max:{None}, min:{None}, value:{c.value}")  

def test_interval(interval,value):
  if not interval is None:
    if np.any([i is None for i in interval]):
      if interval[0] is None:
        value=interval[1]
      else:
        value=interval[0]
      interval=None	
  return interval,value

def append_to_ID(ID,intlist):
  inID=False
  for i in intlist:
    if i in ID:
      inID=True
      break
  if inID:
    for j in intlist:
      if not j in ID:
        ID.append(j)
    return True
  else:
    return False



def normalize(H,include):
  C=-H[include][:,include]
  d=np.maximum(np.diag(C).reshape((len(C),1)),1e-30)**0.5
  C=C/(d*d.T)
  includemap=np.arange(len(include))[include]
  return C,includemap

def decomposition(H,include=None):
  if include is None:
    include=[True]*len(H)
  C,includemap=normalize(H, include)
  c_index, var_prop, d, p = stat.var_decomposition(XXNorm=C)
  c_index=c_index.flatten()
  return c_index, var_prop,includemap


def multicoll_problems(computation,H,include,mc_problems):
  c_index, var_prop, includemap = decomposition(H, include)
  MAX_VAR_PROP_FRACTION = 0.4
  if c_index is None:
    return False,False
  mc_list=[]
  largest_ci=None
  limit = computation.panel.options.multicoll_threshold_report.value
  for cix in range(1,len(c_index)):
    if (np.sum(var_prop[-cix]>MAX_VAR_PROP_FRACTION)>1) and (c_index[-cix]>limit):
      if largest_ci is None:
        largest_ci=c_index[-cix]
      var_prop_ix=np.argsort(var_prop[-cix])[::-1]
      var_prop_val=var_prop[-cix][var_prop_ix]
      j=var_prop_ix[0]
      j=includemap[j]
      done=var_prop_check(computation,var_prop_ix, var_prop_val, includemap,j,mc_problems,c_index[-cix],mc_list)
      if done:
        break
  return c_index[-1],mc_list

def var_prop_check(computation,var_prop_ix,var_prop_val,includemap,assc,mc_problems,cond_index,mc_list):
  i = 1
  if var_prop_val[i]<0.5:
    return True
  index=var_prop_ix[i]
  index=includemap[index]
  mc_problems.append([index,assc,cond_index])
  mc_list.append(index)
  return False

def get_no_check(computation):
  no_check=computation.panel.options.do_not_constrain.value
  X_names=computation.panel.input.X_names
  if not no_check is None:
    if no_check in X_names:
      return X_names.index(no_check)
    print("A variable was set for the 'Do not constraint' option (do_not_constrain), but it is not among the x-variables")

def constraint_multicoll(k,computation,include,mc_problems, H):
  CI_max=0
  for i in range(k-1):
    CI,mc_list=multicoll_problems(computation, H,include,mc_problems)
    CI_max=max((CI_max,CI))
    if len(mc_list)==0:
      break
    include[mc_list]=False
  return CI_max







#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np

class re_obj:
  def __init__(self,panel,group,T_i,T_i_count,fixed_random_eff):
    """Following Greene(2012) p. 413-414"""
    if fixed_random_eff==0:
      self.FE_RE=0
      return
    self.sigma_u=0
    self.group=group
    self.avg_Tinv=np.mean(1/T_i_count) #T_i_count is the total number of observations for each group (N,1)
    self.T_i=T_i*panel.included[3]#self.T_i is T_i_count at each observation (N,T,1)
    self.FE_RE=fixed_random_eff

  def RE(self,x,panel,recalc=True):
    if self.FE_RE==0:
      return np.zeros(x.shape)
    if self.FE_RE==1:
      self.xFE=self.FRE(x,panel)
      return self.xFE
    if recalc:
      N,T,k=x.shape

      self.xFE=(x+self.FRE(x,panel))*panel.included[3]
      self.e_var=panel.mean(self.xFE**2)/(1-self.avg_Tinv)
      self.v_var=panel.mean((panel.included[3]*x)**2)-self.e_var
      if self.v_var<0:
        #print("Warning, negative group random effect variance. 0 is assumed")
        self.v_var=0
        self.theta=panel.zeros[3]
        return panel.zeros[3]
      self.theta=(1-np.sqrt(self.e_var/(self.e_var+self.v_var*self.T_i)))*panel.included[3]
      self.theta*=(self.T_i>1)
      if np.any(self.theta>1) or np.any(self.theta<0):
        raise RuntimeError("WTF")
    #x=panel.Y
    eRE=self.FRE(x,panel,self.theta)
    return eRE

  def dRE(self,dx,x,vname,panel):
    """Returns the first and second derivative of RE"""
    if dx is None:
      return None
    if self.FE_RE==0:
      return np.zeros(dx.shape)		
    panel=panel
    if not hasattr(self,'dxFE'):
      self.dxFE=dict()
      self.dFE_var=dict()
      self.dtheta=dict()
      self.de_var=dict()
      self.dv_var=dict()

    if dx is None:
      return None
    elif self.FE_RE==1:
      return self.FRE(dx,panel)	
    if self.v_var==0:
      return np.zeros(dx.shape)
    (N,T,k)=dx.shape	

    self.dxFE[vname]=(dx+self.FRE(dx,panel))*panel.included[3]
    self.de_var[vname]=2*np.sum(np.sum(self.xFE*self.dxFE[vname],0),0)/(panel.NT*(1-self.avg_Tinv))
    self.dv_var[vname]=(2*np.sum(np.sum(x*dx*panel.included[3],0),0)/panel.NT)-self.de_var[vname]		

    self.dtheta_de_var=(-0.5*(1/self.e_var)*(1-self.theta)*self.theta*(2-self.theta))
    self.dtheta_dv_var=(0.5*(self.T_i/self.e_var)*(1-self.theta)**3)
    self.dtheta[vname]=(self.dtheta_de_var*self.de_var[vname]+self.dtheta_dv_var*self.dv_var[vname])
    self.dtheta[vname]*=(self.T_i>1)
    dRE0=self.FRE(dx,panel,self.theta)
    dRE1=self.FRE(x,panel,self.dtheta[vname])
    ret=(dRE0+dRE1)*panel.included[3]
    remove_extreemes(ret)
    return ret

  def ddRE(self,ddx,dx1,dx2,x,vname1,vname2,panel):
    """Returns the first and second derivative of RE"""
    if self.FE_RE==0:
      return 0*panel.included[4]		
    if dx1 is None or dx2 is None:
      return None
    (N,T,k)=dx1.shape
    (N,T,m)=dx2.shape			
    if self.sigma_u<0:
      return 0*panel.included[4]
    elif self.FE_RE==1:
      return self.FRE(ddx,panel)	
    if self.v_var==0:
      return panel.zeros[4]

    if ddx is None:
      ddxFE=0
      ddx=0
      hasdd=False
    else:
      ddxFE=(ddx+self.FRE(ddx,panel))*panel.included[4]
      hasdd=True

    dxFE1=self.dxFE[vname1].reshape(N,T,k,1)
    dxFE2=self.dxFE[vname2].reshape(N,T,1,m)
    dx1=dx1.reshape(N,T,k,1)
    dx2=dx2.reshape(N,T,1,m)
    de_var1=self.de_var[vname1].reshape(k,1)
    de_var2=self.de_var[vname2].reshape(1,m)
    dv_var1=self.dv_var[vname1].reshape(k,1)
    dv_var2=self.dv_var[vname2].reshape(1,m)		
    dtheta_de_var=self.dtheta_de_var.reshape(N,T,1,1)
    dtheta_dv_var=self.dtheta_dv_var.reshape(N,T,1,1)
    theta=self.theta.reshape(N,T,1,1)
    x=x.reshape(N,T,1,1)
    incl=panel.included[4]

    overflow=False
    theta_args=dxFE1, dxFE2, ddxFE, dx1, dx2, x, ddx, dtheta_dv_var, theta, dtheta_de_var, de_var1, de_var2, dv_var1, dv_var2,panel
    try:
      ddtheta=self.calc_theta(*theta_args)
    except (RuntimeWarning,OverflowError) as e:
      print(e)
      remove_extreemes(theta_args)
      ddtheta=self.calc_theta(*theta_args)
      overflow=True

    if hasdd:
      dRE00=self.FRE(ddx,panel,self.theta.reshape(N,T,1,1))
    else:
      dRE00=0
    dRE01=self.FRE(dx1,panel,self.dtheta[vname2].reshape(N,T,1,m),True)
    dRE10=self.FRE(dx2,panel,self.dtheta[vname1].reshape(N,T,k,1),True)
    dRE11=self.FRE(x,panel,ddtheta,True)
    ret=(dRE00+dRE01+dRE10+dRE11)*panel.included[4]
    if overflow:
      remove_extreemes([ret])
    return (dRE00+dRE01+dRE10+dRE11)*panel.included[4]

  def calc_theta(self,dxFE1,dxFE2,ddxFE,dx1,dx2,x,ddx,dtheta_dv_var,theta,dtheta_de_var,de_var1,de_var2,dv_var1,dv_var2,panel):
    incl=panel.included[4]
    (N,T,k,_)=dx1.shape
    (N,T,_,m)=dx2.shape		
    T_i=self.T_i.reshape(N,T,1,1)

    d2e_var=2*np.sum(np.sum(dxFE1*dxFE2+self.xFE.reshape(N,T,1,1)*ddxFE,0),0)/(panel.NT*(1-self.avg_Tinv))
    d2v_var=(2*np.sum(np.sum((dx1*dx2+x*ddx)*incl,0),0)/panel.NT)-d2e_var	

    d2theta_d_e_v_var=-0.5*dtheta_dv_var*(1/self.e_var)*(3*(theta-2)*theta+2)
    d2theta_d_v_var =-0.75*(T_i/self.e_var)**2*(1-theta)**5
    d2theta_d_e_var =-0.5*dtheta_de_var*(1/self.e_var)*(4-3*(2-theta)*theta)	

    ddtheta  =d2theta_d_e_var  * de_var1* de_var2 
    ddtheta +=d2theta_d_e_v_var * (de_var1* dv_var2+dv_var1* de_var2)
    ddtheta +=d2theta_d_v_var * dv_var1* dv_var2  
    ddtheta +=dtheta_de_var*d2e_var+dtheta_dv_var*d2v_var
    ddtheta*=(T_i>1)	

    return ddtheta

  def FRE(self,x,panel,w=1,d=False):
    if self.group:
      return self.FRE_group(x,w,d,panel)
    else:
      return self.FRE_time(x,w,d,panel)

  def FRE_group(self,x,w,d,panel):
    """returns x after fixed effects, and set lost observations to zero"""
    #assumes x is a "N x T x k" matrix
    if x is None:
      return None
    T_i,s=get_subshapes(panel,x,True)
    incl=panel.included[len(s)]

    sum_x=np.sum(x*incl,1).reshape(s)
    mean_x=sum_x/T_i
    mean_x_all=np.sum(sum_x,0)/panel.NT
    try:
      dFE=w*(mean_x_all-mean_x)*incl#last product expands the T vector to a TxN matrix
    except (RuntimeWarning,OverflowError) as e:
      print(e)
      remove_extreemes([w])
      dFE=w*(mean_x_all-mean_x)*incl#last product expands the T vector to a TxN matrix
      remove_extreemes([dFE])
    return dFE

  def FRE_time(self,x,w,d,panel):
    #assumes x is a "N x T x k" matrix


    if x is None:
      return None
    mean_x,mean_x_all,incl=mean_time(panel, x)
    try:
      dFE=(w*(mean_x_all-mean_x))*incl#last product expands the T vector to a TxN matrix
    except (RuntimeWarning,OverflowError) as e:
      print(e)
      remove_extreemes([w])
      dFE=(w*(mean_x_all-mean_x))*incl#last product expands the T vector to a TxN matrix		
      remove_extreemes([dFE])
    return dFE

def mean_time(panel,x,mean_dates=False):
  n_dates=panel.n_dates
  dmap=panel.date_map
  date_count,s=get_subshapes(panel,x,False)
  incl=panel.included[len(s)]
  x=x*incl
  sum_x_dates=np.zeros(s)
  for i in range(n_dates):
    sum_x_dates[i]=np.sum(x[dmap[i]],0)		
  mean_x_dates=sum_x_dates/date_count
  if mean_dates:
    return mean_x_dates
  mean_x=np.zeros(x.shape)
  for i in range(n_dates):
    mean_x[dmap[i]]=mean_x_dates[i]	
  mean_x_all=np.sum(sum_x_dates,0)/panel.NT
  return mean_x,mean_x_all,incl



def get_subshapes(panel,x,group):
  if group:
    if len(x.shape)==3:
      N,T,k=x.shape
      s=(N,1,k)
      T_i=panel.T_i
    elif len(x.shape)==4:
      N,T,k,m=x.shape
      s=(N,1,k,m)
      T_i=panel.T_i.reshape((N,1,1,1))	
    return T_i,s
  else:
    date_count=panel.date_count
    n_dates=panel.n_dates
    if len(x.shape)==3:
      N,T,k=x.shape
      s=(n_dates,1,k)

    elif len(x.shape)==4:
      N,T,k,m=x.shape
      s=(n_dates,1,k,m)
      date_count=date_count.reshape((n_dates,1,1,1))	
    return date_count,s			



def remove_extreemes(args,max_arg=1e+100):
  for i in range(len(args)):
    s=np.sign(args[i][np.abs(args[i])>max_arg])
    args[i][np.abs(args[i])>max_arg]=s*max_arg


#!/usr/bin/env python
# -*- coding: utf-8 -*-


from . import parallel
from . import main
from . import options as opt_module
from . import output
from .processing import loaddata
from . import debug
from . import info


import numpy as np
import os
import sys
import time
try:
  import matplotlib
except:
  matplotlib = None
import pandas as pd

import inspect


mp = None

CALLBACK_ACTIVE = True


def enable_parallel():
  global mp
  N_NODES = 1
  PARALLEL = True #change to false for debugging

  t0=time.time()

  #temporary debug output is saved here:

  mp = parallel.Parallel(N_NODES, PARALLEL, CALLBACK_ACTIVE)

  mp.exec("from paneltime import maximization\n", 'init')

  print(f"parallel: {time.time()-t0}")


def execute(model_string,dataframe, ID=None,T=None,HF=None,instruments=None, console_output=True):

  """Maximizes the likelihood of an ARIMA/GARCH model with random/fixed effects (RE/FE)\n
	model_string: a string on the form 'Y ~ X1 + X2 + X3\n
	dataframe: a dataframe consisting of variables with the names usd in model_string, ID, T, HF and instruments\n
	ID: The group identifier\n
	T: the time identifier\n
	HF: list with names of heteroskedasticity factors (additional regressors in GARCH)\n
	instruments: list with names of instruments
	console_output: if True, GUI output is turned off (GUI output is experimental)
	"""

  window=main.identify_global(inspect.stack()[1][0].f_globals,'window', 'geometry')
  exe_tab=main.identify_global(inspect.stack()[1][0].f_globals,'exe_tab', 'isrunning')

  r=main.execute(model_string,dataframe,ID, T,HF,options,window,exe_tab,instruments, console_output, mp)

  return r

def load_json(fname):

  if False:#detects previously loaded dataset in the environment
    dataframe=main.indentify_dataset(globals(),fname)
    if (not dataframe==False) and (not dataframe is None):
      return dataframe	
  try:
    dataframe=main.loaddata.load_json(fname)
  except FileNotFoundError:
    raise RuntimeError("File %s not found" %(fname))
  return dataframe


def load(fname,sep=None):

  """Loads data from file <fname>, asuming column separator <sep>.\n
	Returns a dataframe (a dictionary of numpy column matrices).\n
	If sep is not supplied, the method will attemt to find it."""
  if False:#detects previously loaded dataset in the environment
    dataframe=main.indentify_dataset(globals(),fname)
    if (not dataframe==False) and (not dataframe is None):
      return dataframe	
  try:
    dataframe=main.loaddata.load(fname,sep)
  except FileNotFoundError:
    raise RuntimeError("File %s not found" %(fname))
  return dataframe

def load_SQL(conn,sql_string):

  """Loads data from an SQL server, using sql_string as query"""
  if False:#detects previously loaded dataset in the environment
    dataframe=main.indentify_dataset(globals(),sql_string)
    if (not dataframe==False) and (not dataframe is None):
      return dataframe
  dataframe=main.loaddata.load_SQL(sql_string,conn)
  #except RuntimeError as e:
  #	raise e
  return dataframe

version = info.version

options=opt_module.regression_options()
preferences=opt_module.application_preferences()





#!/usr/bin/env python
# -*- coding: utf-8 -*-

DEFAULT_INTERCEPT_NAME='Intercept'
VAR_INTERCEPT_NAME='log_variance_constant'
INSTRUMENT_INTERCEPT_NAME='instrument_intercept'
CONST_NAME='one'
NUMERIC_TAG="_numeric"

import numpy as np
import pandas as pd

def get_variables(ip,df,model_string,IDs,timevar,heteroscedasticity_factors,instruments,settings,pool=(None,'mean')):
  if not settings.supress_output:
    print ("Analyzing variables ...")
  if not type(df)==pd.DataFrame:
    raise RuntimeError('The dataframe supplied is not a pandas dataframe. Only pandas dataframes are supported.')
  if CONST_NAME in df:
    print(f"Warning: The name {CONST_NAME} is reserved for the constant 1."
                      f"The variable with this name will be overwritten and set to 1")
  df=pool_func(df,pool)

  df[CONST_NAME]=1
  df[DEFAULT_INTERCEPT_NAME]     = df[CONST_NAME]
  df[VAR_INTERCEPT_NAME]         = df[CONST_NAME]
  df[INSTRUMENT_INTERCEPT_NAME]  = df[CONST_NAME]

  if IDs is None:
    IDs = 'IDs'
    df[IDs] = 1
  if timevar is None:
    timevar = 'timevar'
    df[timevar] = np.arange(len(df))
    
  IDs     = get_names(IDs, 'IDs')
  timevar =get_names(timevar, 'timevar')

  IDs_num,     IDs      = handle_IDs(ip,df,IDs)
  timevar_num, timevar  = handle_time(ip,df,timevar)	
  sort= IDs_num + timevar
  if len(sort):
    df=df.sort_values(sort)
  pd_panel = df
  if len(IDs_num)>0:
    pd_panel=df.groupby(IDs_num)

  W=get_names(heteroscedasticity_factors,'heteroscedasticity_factors',True, VAR_INTERCEPT_NAME)
  Z=get_names(instruments,'instruments',True,INSTRUMENT_INTERCEPT_NAME)

  try:
    Y,X=parse_model(model_string, settings)
  except:
    raise RuntimeError("The model_string must be on the form Y~X1+X2+X3")	
  if Y==['']:
    raise RuntimeError("No dependent variable specified")		

  x = IDs_num+IDs+timevar+timevar_num+W+Z+Y+X
  x = list(dict.fromkeys(x))
  df,ip.max_lags=eval_variables(x, df, pd_panel)
  df=df[x]
  n=len(df)
  df=df.dropna()
  ip.lost_na_obs=(n-len(df))-ip.max_lags
  const={}
  for x,add_intercept,num in [
          ('IDs_num',False,True),('timevar_num',False,True),
                                ('IDs',False,False),('timevar',False,False),
                                ('W',True,True),('Z',True,True),('Y',False,True),
                                  ('X',settings.add_intercept.value,True)]:
    ip.__dict__[x], const[x]= check_var(df,locals()[x],x,add_intercept,num)
    ip.__dict__[x+"_names"]=list(ip.__dict__[x].columns)
  ip.dataframe=df
  ip.has_intercept=const['X']



def pool_func(df,pool):
  x,operation=pool
  if x is None:
    return df
  x=get_names(x, 'pool')
  df=df.groupy(x).agg(operation)
  return df

def check_var(df,x,inputtype,add_intercept,numeric):
  if len(x)==0:
    return None,None
  dfx=df[x]
  if not numeric:
    return dfx,None
  const_found=False
  for i in x:
    if ' ' in i:
      raise RuntimeError(f'Spaces are not allowed in variables, but found in the variable {i} from {inputtype}')
    try:
      v=np.var(dfx[i])
    except TypeError as e:
      raise TypeError(f"All variables except time and id must be numeric. {e}")
    if v==0 and const_found:
      if dfx[i].iloc[0]==0:
        print(f"Warning: All values in {i} from {inputtype} are zero, variable dropped")
      else:
        print(f"Warning: {i} from {inputtype} is constant. Variable dropped.")
      dfx=dfx.drop(i,1)
    elif v==0 and not const_found:
      if inputtype=='Y':
        raise RuntimeError('The dependent variable is constant')	
      const_found=True
  return dfx,const_found

def eval_variables(x,df,pd_panel):
  lag_obj=lag_object(pd_panel)
  d={'D':lag_obj.diff,'L':lag_obj.lag,'np':np}	
  for i in df.keys():
    d[i]=df[i]
  for i in x:
    df[i]=eval(i,d)
  return df,lag_obj.max_lags

class lag_object:
  def __init__(self,panel):
    self.panel=panel
    self.max_lags=0

  def lag(self,variable,lags=1):
    x=self.panel[variable.name].shift(lags)
    self.max_lags=max((self.max_lags,lags))
    return x

  def diff(self,variable,lags=1):
    x=self.panel[variable.name].diff(lags)
    self.max_lags=max((self.max_lags,lags))
    return x

def parse_model(model_string,settings):
  split = None
  for i in ['~','=']:
    if i in model_string:
      split=i
      break
  if split is None:#No dependent
    return [model_string],[DEFAULT_INTERCEPT_NAME]
  Y,X=model_string.split(split)
  X=[i.strip() for i in X.split('+')]
  if X==['']:
    X=[DEFAULT_INTERCEPT_NAME]
  if settings.add_intercept.value and not (DEFAULT_INTERCEPT_NAME in X):
    X=[DEFAULT_INTERCEPT_NAME]+X
  return [Y],X

def get_names(x,inputtype,add_intercept=False,intercept_name=None):
  if x is None:
    r=[]
  elif type(x)==str:
    r=[x]
  elif type(x)==pd.DataFrame:		
    r=x.columns
  elif type(x)==pd.Series:
    r=[x.name]
  else:
    raise RuntimeError(f"Input for {inputtype} needs to be a list or tuple of strings, a pandas DataFrame object or a pandas Series object")
  if add_intercept:
    r=[intercept_name]+r
  return r

def handle_time(ip,df,x):
  if x==[]:
    return [],[]
  x=x[0]
  try:
    x_dt=pd.to_datetime(df[x])
  except ValueError as e:
    try:
      x_dt=pd.to_numeric(x_dt)
    except ValueError as e:
      raise ValueError(f'Expected date or numeric for {inputtype}, but {x} is not recognized as a date or numeric variable by pandas.')
  x_dt=pd.to_numeric(x_dt)/(24*60*60*1000000000)
  x_int=x_dt.astype(int)
  if len(pd.unique(df[x]))==len(pd.unique(x_int)):
    x_dt=x_int
  df[x+NUMERIC_TAG]=x_dt
  return [x+NUMERIC_TAG],[x]


def handle_IDs(ip,df,x):
  if x==[]:
    return [],[]
  x=x[0]
  ids, ip.IDs_unique = pd.factorize(df[x],True)
  df[x+NUMERIC_TAG]=ids
  #both these are true before next assignment:
  #np.all(ip.IDs_unique[ids]==df[x])
  #np.all(np.arange(len(ids))[ids]==ids)
  return [x+NUMERIC_TAG],[x]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np

def save_csv(fname, array, sep = ','):
  f = open(fname, 'wt')
  try:
    for line in array:
      f.write(sep.join(line))
  except TypeError as e:
    np.savetxt(fname,array,fmt='%s', delimiter=sep)
  f.close()


def dot(a,b,reduce_dims=True):
  """Matrix multiplication. Returns the dot product of a*b where either a or be or both to be
  arrays of matrices. Faster than mmult, less general and only used for special purpose.
  Todo: generalize and merge"""


  if len(a.shape)==3 and len(b.shape)==2:
    x = np.array([np.dot(a[i],b) for i in range(a.shape[0])])
  elif len(a.shape)==3 and len(b.shape)==3:
    x = np.sum([np.dot(a[i].T,b[i]) for i in range(a.shape[0])],0)
  elif len(a.shape)==2 and len(b.shape)==3:
    x = np.array([np.dot(a,b[i]) for i in range(b.shape[0])])
  return x


class ArmaDot:
  def __init__(self):
    pass

  def dotroll(self,aband,k,sign,b,ll):
    x = sign*self.fast_dot(aband, b)
    w=[]
    for i in range(k):
      w.append(np.roll(np.array(x),i+1,1))
      w[i][:,:i+1]=0
    x=np.array(w)
    x=np.moveaxis(x,0,2)
    return x


  def fast_dot(self, a, b):
    a_, name = a
    n = get_n(a_)
    if n is None:
      n = len(a_)

    r = a_[0]*b
    for i in range(1,n):
      r[:,i:] += a_[i]*b[:,:-i]

    return r


  def dot(self,a,b,ll):
    if len(a)>2:#then this is a proper matrix
      (aband,k,sgn)=a
      if k==0:
        return None
      return self.dotroll(aband, k, sgn, b, ll)
    x = self.fast_dot(a, b)
    return x


def get_n(a):
  minval = 0
  a_1 = np.abs(a[1:])
  max_a = np.max(a_1)
  if np.min(np.abs(a_1)) >= minval:
    return None
  if max_a == 0:
    return 1		
  else:
    nz = np.nonzero(a_1/max_a < minval)[0]
    if len(nz)>0:
      return nz[0]+1
    else:
      return None


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from . import stat_functions as stat
from . import stat_dist

import numpy as np
import os
try:
  from matplotlib import pyplot  as plt
except:
  plt = None



class ProcessCharts():
  def __init__(self,panel):
    if plt is None:
      return
    self.panel=panel
    self.ll=None	
    self.subplot=plt.subplots(1,figsize=(4,2.5),dpi=75)
    self.chart_list=[
                  ['histogram',self.histogram],
                        ['correlogram',self.correlogram],
                        ['correlogram_variance',self.correlogram_variance]
                ]

  def save_all(self,ll):
    self.ll=ll
    if self.ll is None:
      return
    for name,chart in self.chart_list:
      chart(f'img/{name}.png')				

  def histogram(self,f):
    N,T,k=self.panel.X.shape
    fgr,axs = self.subplot
    n=self.ll.e_norm_centered.shape[2]
    e=self.ll.e_norm_centered[self.panel.included[2]].flatten()
    N=e.shape[0]
    e=e.reshape((N,1))

    grid_range=4
    grid_step=0.05	
    h,grid=histogram(e,grid_range,grid_step)
    norm=stat_dist.norm(grid, cdf = False)*grid_step	

    axs.bar(grid,h,color='grey', width=0.025,label='histogram')
    axs.plot(grid,norm,'green',label='normal distribution')
    axs.legend(prop={'size': 6})
    name='Histogram - frequency'
    axs.set_title(name)
    self.save(f)

  def correlogram(self,f):
    fgr,axs=self.subplot
    lags=20
    rho=stat.correlogram(self.panel, self.ll.e_norm_centered,lags)
    x=np.arange(lags+1)
    axs.bar(x,rho,color='grey', width=0.5,label='correlogram')
    name='Correlogram - residuals'
    axs.set_title(name)
    self.save(f)

  def correlogram_variance(self,f):
    N,T,k=self.panel.X.shape
    fgr,axs=self.subplot
    lags=20
    e2=self.ll.e_norm_centered**2
    e2=(e2-self.panel.mean(e2))*self.panel.included[3]
    rho=stat.correlogram(self.panel, e2,lags)
    x=np.arange(lags+1)
    axs.bar(x,rho,color='grey', width=0.5,label='correlogram')
    name='Correlogram - squared residuals'
    axs.set_title(name)
    self.save(f)

  def save(self,save_file):
    fgr,axs=self.subplot
    fgr.savefig(save_file)
    axs.clear()	

def histogram(x,grid_range,grid_step):
  N,k=x.shape
  grid_n=int(2*grid_range/grid_step)
  grid=np.array([i*grid_step-grid_range for i in range(grid_n)]).reshape((1,grid_n))
  ones=np.ones((N,1))
  x_u=np.concatenate((ones,x>=grid),1)
  x_l=np.concatenate((x<grid,ones),1)
  grid=np.concatenate((grid.flatten(),[grid[0,-1]+grid_step]))
  histogram=np.sum((x_u*x_l),0)
  if int(np.sum(histogram))!=N:
    raise RuntimeError('Error in histogram calculation')
  return histogram/N,grid





#!/usr/bin/env python
# -*- coding: utf-8 -*-

from . import linesearch
from ..parallel import callback
from . import direction

import numpy as np
import time

#This module finds the array of arguments that minimizes some function. The derivative 
#of the function also needs to be supplied. 
#This is an adaption of the Broyden-Fletcher-Goldfarb-Shanno variant of Davidon-Fletcher-Powell algorithm by 
#Press, William H,Saul A Teukolsky,William T Vetterling and Brian P Flannery. 1992. Numerical Recipes in C'. 
#Cambridge: Cambridge University Press.



EPS=3.0e-16 
TOLX=(4*EPS) 
GTOL = 1e-5

def dfpmax(x, comput, callback, panel, slave_id):
  """Given a starting point x[1..n] that is a vector of length n, the Broyden-Fletcher-Goldfarb-
  Shanno variant of Davidon-Fletcher-Powell minimization is performed on a function func, using
  its gradient as calculated by a routine dfunc. The convergence requirement on zeroing the
  gradient is input as gtol. Returned quantities are x[1..n] (the location of the minimum),
  iter (the number of iterations that were performed), and fret (the minimum value of the
  function). The routine lnsrch is called to perform approximate line minimizations.
  fargs are fixed arguments that ar not subject to optimization. ("Nummerical Recipes for C") """

  x, ll, f, g, hessin, H = comput.calc_init_dir(x)

  its, msg = 0, ''
  MAXITER = 10000

  cbhandler = CallBackHandler(callback, slave_id)
  fdict = {}
  for its in range(MAXITER):  	#Main loop over the iterations.


    dx, dx_norm, H_ = direction.get(g, x, H, comput.constr, f, hessin, comput.ev_constr, simple=False)
    ls = linesearch.LineSearch(x, comput, panel, ll)
    ls.lnsrch(x, f, g, H, dx)	

    dx_realized = ls.x - x
    incr = ls.f - f
    fdict[its] = ls.f
    ll = ls.ll


    x, f, hessin, H, g, conv, se, det, anal = comput.exec(dx_realized,  hessin, H, incr, its, ls)

    err = (np.max(np.abs(dx_realized)) < TOLX) and its>5

    terminate = (conv>0) or err or its+1==MAXITER

    #print(f"sid:{slave_id}, f:{ls.f}, conv:{conv}, its:{its}")

    if conv==1:
      msg = "Convergence on zero gradient; local or global minimum identified"
    elif conv==2:
      msg = "Convergence on zero expected gain; local or global minimum identified given multicolinearity constraints"		
    elif conv==3:
      msg = "Reached the maximum number of iterations"		  
    elif err:
      msg = "Warning: Convergence on delta x; the gradient is incorrect or the tolerance is set too low"
    elif terminate:
      msg = "No convergence within %s iterations" %(MAXITER,)


    cbhandler.assign(ls, msg, dx_norm, f, x, H, comput, g, hessin, dx_realized, 
                                 incr, its, 'linesearch', terminate, 
                                                  conv, fdict, se, det)			

    if terminate or cbhandler.quit:	
      if terminate:
        cause = msg
      else:
        cause = 'forced'
      if not panel.options.supress_output:
        print(f"quit slave {slave_id}, time: {time.time()}, cause: {cause}, conv:{conv}")
      return cbhandler.callback.outbox, ls.ll

class CallBackHandler:
  def __init__(self, callback, slave_id):
    self.t = time.time()
    self.callback = callback
    self.id = slave_id
    self.quit = False
    self.inbox = {}


  def assign(self, ls, msg, dx_norm, f, x, H, comput, g, hessin, dx, incr, its, 
                   task, terminate, conv, fdict, se, det):


    self.check_for_quit_order()



    if msg == '':
      msg = ls.msg	

    if time.time()-self.t<1 and not (self.quit or terminate):
      return
    self.t = time.time()

    self.inbox = self.callback.callback(msg = msg, dx_norm = dx_norm, f = f, x = x, 
                                                    H = H, G=comput.G, g = g, hessin = hessin, dx = dx, 
                                 incr = incr, rev = ls.rev, alam = ls.alam, 
                                 its = its, constr = comput.constr, perc=min(its/100, 1), task = task, 
                                 terminate= terminate or self.quit, conv = conv, slave_id = self.id, 
                                 fdict = dict(fdict), CI = comput.CI, CI_anal = comput.CI_anal, tpgain = comput.totpgain, se = se, det = det)

  def check_for_quit_order(self):

    if not 'quit' in  self.callback.inbox:
      return
    if not self.callback.inbox['quit']:
      return
    self.quit = True








#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
  from . import callback
except:
  import callback


import sys
import os
import traceback
import datetime
import pickle
import gc
import inspect
import time
from threading import Thread
import subprocess
from queue import Queue
from queue import Empty



class Slave:
  def __init__(self, t, s_id, path, print_to_file, n_nodes):
    self.t = t
    self.id = s_id
    self.n_nodes = n_nodes
    self.d = dict()
    self.d['slave_id'] = s_id
    self.processes = Processes(self)
    while 1:
      self.wait_for_orders()


  def wait_for_orders(self):

    (msg,obj) = self.t.receive()
    response = None
    if msg == 'kill':
      sys.exit()
      response = True

    elif msg == 'callback':
      name, d = obj
      if name in self.processes:
        response = dict(self.processes[name].transact(d))
        response['slave_id'] = self.id
      else:
        response = {'empty':True}		

    elif msg == 'check_state':
      response = self.processes.countalive(obj)
    elif msg == 'collect':
      response = self.processes.collect(obj)		
    else:#starting a new procedure
      self.processes[msg] = obj
    self.t.send(response)	
    gc.collect()



class Processes(dict):

  def __init__(self,slave):
    self.d = slave.d
    self.parent = slave

  def __setitem__(self, name, task):
    super().__setitem__(name, Process(name, task, self))

  def countalive(self, name = None):
    if not name is None:
      if name in self:
        return self[name].thread.is_alive()
      return 0
    count = 0
    for k in self:
      if hasattr(self[k],'thread'):
        if self[k].thread.is_alive():
          count+=1
    return count

  def collect(self, name):
    if not name in self:
      return None, None
      #raise RuntimeError(f"The process '{name}' does not exist in node {self.parent.id}")
    d, r = self[name].collect()
    for k in d:
      self.d[k] = d[k]
    return r




class Process:
  def __init__(self, name, task, parent):			
    self.name = name
    self.parent = parent
    self.task = task
    self.d = dict(parent.d)
    self.d['inbox'] = {}
    self.d['outbox'] = {}
    self.threadq = Queue()
    self.inbox = self.d['outbox']
    self.outbox = self.d['inbox']
    self.result = {}, None

    if name == 'transfer dictionary':
      t0=time.time()
      self.get_dict(task)
    else:
      if name == 'debug':
        execute_task(self.d, task, self.threadq)
      else:
        prtask = task.replace('\n', ' ')
        #print(f"id inbox: {id(self.inbox)}, id outbox:{id(self.outbox)}, task:{prtask}")
        self.thread = Thread(target=execute_task, args=(self.d, task, self.threadq, self.name))
        self.thread.start()	
    a=0

  def get_dict(self, fname):	#Blocking	
    try:
      f = open(fname,'rb')
      d = None
      d =  pickle.load(f)
      f.close()	
      for k in d:
        self.parent.d[k] = d[k]

    except Exception as e:
      traceback.print_exc(file = sys.stdout)
      raise e

  def transact(self, incoming):
    #print(f'reading from {id(self.inbox)}')
    if not self.thread.is_alive():
      return dict(self.inbox)
    for k in incoming:
      self.outbox[k] = incoming[k]	
    return dict(self.inbox)


  def collect(self):
    if self.threadq.empty():
      return self.result
    if not self.threadq.get()=='start':
      raise RuntimeError("Expected a start signal")
    d, r = self.threadq.get()
    self.result = d, r
    return d, r




def execute_task(d, task, queue,name):#None blocking 
  queue.put('start')
  d = dict(d)
  try:
    r = execute_task_(d, task)			
  except Exception as e:
    traceback.print_exc(file = sys.stdout)
    raise e
  queue.put((d,r))


def execute_task_(d, task):
  try:#If a returned value is desired, the task must be a function call
    return eval(task, globals(), d)
  except SyntaxError as e:#else, it is assumed that the script only generates local variables for later use
    exec(task, globals(), d)

def run(transact, print_to_file = True):

  if not print_to_file:
    msg, (s_id, path, n_nodes), fname = handshake(transact, print_to_file)
    #Wait for instructions:
    Slave(transact, s_id, path, print_to_file, n_nodes)			
    return
  try: 
    msg, (s_id, path, n_nodes), fname = handshake(transact, print_to_file)
    f = open(fname,'w', 1)
    f.close()
    #Wait for instructions:		
    fstdout = os.path.join(path,f'slaves/{s_id}.txt')
    if print_to_file:
      sys.stdout = open(fstdout,'w',1)
    Slave(transact, s_id, path, print_to_file, n_nodes)
  except Exception as e:
    print(f'error was: {e}')
    try:
      traceback.print_exc(file=sys.stdout)
    except Exception as e:
      print(e)
    f.write('SID: %s      TIME:%s \n' %(s_id,datetime.datetime.now()))
    traceback.print_exc(file = f)
    f.flush()
    f.close()
    raise e

def handshake(transact, print_to_file):
  #Handshake:
  transact.send(os.getpid())
  path='.'
  msg, (s_id, path, n_nodes)=transact.receive()	
  transact.set('slave', n_nodes, path, s_id)
  #error handling
  fname = os.path.join(path,'slave_errors.txt')	
  return msg, (s_id, path, n_nodes), fname


def write(f,txt):
  f = open(f, 'w', 1)
  f.write(str(txt))
  f.flush()
  f.close()	


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module calculates statistics and saves it to a file

from . import stat_functions as stat
from . import stat_dist


import numpy as np
import time

STANDARD_LENGTH=8



class output:
  def __init__(self,ll,panel,comm, dx_norm):
    self.ll=ll
    self.comm=comm
    self.panel=panel
    self.lags=self.panel.options.robustcov_lags_statistics.value[1]
    self.n_variables=self.panel.args.n_args
    self.incr=0
    self.dx_norm = dx_norm
    self.d={'names':np.array(self.panel.args.caption_v),
                        'count':range(self.n_variables),
                                'args':self.ll.args.args_v}
    self.update(comm, 0, ll, self.incr, dx_norm)
    self.heading()

  def update(self,comm, its, ll,incr, dx_norm):
    self.comm=comm
    self.iterations=its
    self.ll=ll
    self.dx_norm = dx_norm
    self.incr=incr
    self.d['args']=ll.args.args_v
    self.constraints_printout()
    self.t_stats()
    self.heading()

  def statistics(self):
    return Statistics(self.ll,self.panel)

  def reg_table(self):
    return RegTableObj(self)

  def t_stats(self):
    d=self.d
    comm=self.comm
    panel=self.panel

    T=len(d['names'])
    if comm.H is None:
      return
    d['se_robust'],d['se_st']=sandwich(comm,self.lags)
    d['se_robust_oposite'],d['se_st_oposite']=sandwich(comm,self.lags,oposite=True)
    d['se_robust'][np.isnan(d['se_robust'])]=d['se_robust_oposite'][np.isnan(d['se_robust'])]
    d['se_st'][np.isnan(d['se_st'])]=d['se_st_oposite'][np.isnan(d['se_st'])]
    #d['se_robust_fullsize'],d['se_st_fullsize']=sandwich(comm,self.lags,resize=False)
    no_nan=np.isnan(d['se_robust'])==False
    valid=no_nan
    valid[no_nan]=(d['se_robust'][no_nan]>0)
    d['tstat']=np.array(T*[np.nan])
    d['tsign']=np.array(T*[np.nan])
    d['tstat'][valid]=d['args'][valid]/d['se_robust'][valid]
    d['tsign'][valid]=(1-stat_dist.tcdf(np.abs(d['tstat'][valid]),panel.df))#Two sided tests
    d['sign_codes']=get_sign_codes(d['tsign'])

  def heading(self):
    CI ='None'
    n_CI = 'None'
    if not self.comm.constr is None:
      if not self.comm.constr.CI is None:
        CI = np.round(self.comm.constr.CI)
      n_CI=len(self.comm.constr.mc_problems)


    self.model_desc =  f"Model:\t{self.panel.input.Y_names[0]} = {' + '.join(self.panel.input.X_names)}\n"
    instr='None'
    if not self.panel.input.Z_names is None:
      self.model_desc += f"Instruments:\t{', '.join(self.panel.input.Z_names[1:])}\n" 

    n,T,k=self.panel.X.shape
    run_time = np.round(time.time()-self.comm.start_time)

    s = (
                  f"LL:\t{self.ll.LL        }\tIteration:\t{self.iterations        }\tPanel observations:\t{self.panel.NT_before_loss }\n"
                f"Increment:\t{self.incr  }\tCollinearity - max(cond.idx):\t{CI  }\tPanel dates:\t{T                                }\n"
                f"Time elapsed:\t{run_time}\tCollinearity - # variables:\t{n_CI  }\tPanel groups:\t{n                               }\n"


                )

    self.output_stats=s


  def constraints_printout(self):
    panel=self.panel
    comm=self.comm
    constr=comm.constr
    weak_mc_dict={}
    if not comm.constr is None:
      weak_mc_dict=comm.constr.weak_mc_dict
    d=self.d
    if not self.dx_norm is None:
      d['dx_norm']=self.dx_norm
    T=len(d['names'])
    d['set_to'],d['assco'],d['cause'],d['multicoll']=['']*T,['']*T,['']*T,['']*T
    if constr is None:
      return
    c=constr.fixed
    for i in c:
      d['set_to'][i]=c[i].value_str
      d['assco'][i]=c[i].assco_name
      d['cause'][i]=c[i].cause	

    c=constr.intervals
    for i in c:
      if not c[i].intervalbound is None:
        d['set_to'][i]=c[i].intervalbound
        d['assco'][i]='NA'
        d['cause'][i]=c[i].cause		

    for i in weak_mc_dict:#adding associates of non-severe multicollinearity
      d['multicoll'][i]='|'
      d['assco'][i]=panel.args.caption_v[weak_mc_dict[i][0]]	


class RegTableObj(dict):
  def __init__(self,output):
    dict.__init__(self)
    self.d=output.d
    self.Y_names=output.panel.input.Y_names
    self.X_names=output.panel.input.X_names
    self.args=output.ll.args.dict_string
    self.n_variables=output.n_variables
    self.output_stats = output.output_stats
    self.model_desc = output.model_desc
    self.footer=f"\n\nSignificance codes: '=0.1, *=0.05, **=0.01, ***=0.001,    |=collinear\n\n{output.ll.err_msg}"	

  def table(self,n_digits = 3,brackets = '(',fmt = 'NORMAL',stacked = True, show_direction = False, show_constraints = True):
    include_cols,llength=self.get_cols(stacked, show_direction, show_constraints)
    if fmt=='INTERNAL':
      self.X=None
      return str(self.args),None
    self.include_cols=include_cols
    self.n_cols=len(include_cols)		
    for a, l,is_string,name,neg,just,sep,default_digits in pr:		
      self[a]=column(self.d,a,l, is_string, name, neg, just, sep, default_digits,self.n_variables)
    self.X=self.output_matrix(n_digits,brackets)
    s=format_table(self.X, include_cols,fmt,
                               "Paneltime GARCH-ARIMA panel regression",
                                           self.model_desc, 
                                           self.output_stats,
                                           self.footer)
    return s,llength


  def output_matrix(self,digits,brackets):
    structured=False
    for i in range(self.n_cols):
      if type(self.include_cols[i])==list:
        structured=True
        break
    if structured:
      return self.output_matrix_structured(digits, brackets)
    else:
      return self.output_matrix_flat(digits, brackets)


  def output_matrix_structured(self,digits,brackets):
    X=[['']*self.n_cols for i in range(3*(self.n_variables+1)-1)]
    for i in range(self.n_cols):
      a=self.include_cols[i]
      if type(a)==list:
        h=self[a[0]].name.replace(':',' ')
        if brackets=='[':
          X[0][i]=f"{h}[{self[a[1]].name}]:"
        elif brackets=='(':
          X[0][i]=f"{h}({self[a[1]].name}):"
        else:
          X[0][i]=f"{h}/{self[a[1]].name}:"
        v=[self[a[j]].values(digits) for j in range(3)]
        for j in range(self.n_variables):
          X[(j+1)*3-1][i]=v[0][j]
          if brackets=='[':
            X[(j+1)*3][i]=f"[{v[1][j]}]{v[2][j]}"
          elif brackets=='(':
            X[(j+1)*3][i]=f"({v[1][j]}){v[2][j]}"
          else:
            X[(j+1)*3][i]=f"{v[1][j]}{v[2][j]}"
      else:
        X[0][i]=self[a].name
        v=self[a].values(digits)
        for j in range(self.n_variables):
          X[(j+1)*3-1][i]=v[j]
    return X	

  def output_matrix_flat(self,digits,brackets):
    X=[['']*self.n_cols for i in range(self.n_variables+1)]
    for i in range(self.n_cols):
      a=self.include_cols[i]
      X[0][i]=self[a].name
      v=self[a].values(digits)
      for j in range(self.n_variables):
        X[j+1][i]=v[j]
    return X	


  def get_cols(self,stacked,
                     show_direction,
                                show_constraints):
    "prints a single regression"
    dx_col=[]
    llength=9
    if show_direction:
      dx_col=['dx_norm']
    else:
      llength-=1
    mcoll_col=[]
    if show_constraints:
      mcoll_col=[ 'multicoll','assco','set_to', 'cause']
    else:
      llength-=2		
    if stacked:
      cols=['count','names', ['args','se_robust', 'sign_codes']] + dx_col + ['tstat', 'tsign'] + mcoll_col
    else:
      cols=['count','names', 'args','se_robust', 'sign_codes'] + dx_col + ['tstat', 'tsign'] + mcoll_col		
    return cols,llength


class column:
  def __init__(self,d,a,l,is_string,name,neg,just,sep,default_digits,n_variables):		
    self.length=l
    self.is_string=is_string
    self.name=name
    self.default_digits=default_digits
    self.neg_allowed=neg
    self.justification=just
    self.tab_sep=sep
    self.n_variables=n_variables
    if a in d:
      self.exists=True
      self.input=d[a]
    else:
      self.exists=False
      self.input=[' - ']*self.n_variables		

  def values(self,digits):
    try:
      if self.length is None:
        if digits=='SCI':
          return self.input
        else:
          return np.round(self.input,digits)
      return np.round(self.input,self.length)
    except:
      if self.length is None:
        return self.input
      else:
        return np.array([str(i).ljust(self.length)[:self.length] for i in self.input])

def sandwich(comm,lags,oposite=False,resize=True):
  panel=comm.panel
  H,G,idx=reduce_size(comm,oposite,resize)
  lags=lags+panel.lost_obs
  onlynans = np.array(len(idx)*[np.nan])
  try:
    hessin=np.linalg.inv(-H)
  except np.linalg.LinAlgError as e:
    print(e)
    return np.array(onlynans),np.array(onlynans)
  se_robust,se,V=stat.robust_se(panel,lags,hessin,G)
  se_robust,se,V=expand_x(se_robust, idx),expand_x(se, idx),expand_x(V, idx,True)
  if se_robust is None:
    se_robust = np.array(onlynans)
  if se is None:
    se = np.array(onlynans)
  return se_robust,se

def reduce_size(comm,oposite,resize):
  #this looks unneccessary complicated
  if comm.constr is None:
    return comm.H, comm.G, np.ones(len(comm.g),dtype=bool)
  H=comm.H
  G=comm.G
  if (G is None) or (H is None):
    return
  m=len(H)
  if not resize:
    return H,G,np.ones(m,dtype=bool)
  weak_mc_dict=comm.constr.weak_mc_dict.keys()
  constr=list(comm.constr.fixed.keys())	
  if oposite:
    weak_mc_dict=[comm.constr.weak_mc_dict[i][0] for i in comm.constr.weak_mc_dict]
    constr=[]
    for i in comm.constr.fixed:
      if not comm.constr.fixed[i].assco_ix is None:
        constr.append(comm.constr.fixed[i].assco_ix)
  if False:#not sure why this is here
    for i in weak_mc_dict:
      if not i in constr:
        constr.append(i)
  idx=np.ones(m,dtype=bool)
  if len(constr)>0:#removing fixed constraints from the matrix
    idx[constr]=False
    H=H[idx][:,idx]
    G=G[:,:,idx]
  return H,G,idx

def expand_x(x,idx,matrix=False):
  x = np.real(x)
  m=len(idx)
  if matrix:
    x_full=np.zeros((m,m))
    x_full[:]=np.nan
    ref=np.arange(m)[idx]
    for i in range(len(x)):
      try:
        x_full[ref[i],idx]=x[i]
        x_full[idx,ref[i]]=x[i]
      except:
        a=0
  else:
    x_full=np.zeros(m)
    x_full[:]=np.nan
    x_full[idx]=x
  return x_full


def get_sign_codes(tsign):
  sc=[]
  for i in tsign:
    if np.isnan(i):
      sc.append(i)
    elif i<0.001:
      sc.append('***')
    elif i<0.01:
      sc.append('** ')
    elif i<0.05:
      sc.append('*  ')
    elif i<0.1:
      sc.append("'  ")
    else:
      sc.append('')
  sc=np.array(sc,dtype='<U3')
  return sc

def remove_illegal_signs(name):
  illegals=['#', 	'<', 	'$', 	'+', 
                  '%', 	'>', 	'!', 	'`', 
                  '&', 	'*', 	'‘', 	'|', 
                  '{', 	'?', 	'“', 	'=', 
                  '}', 	'/', 	':', 	
                  '\\', 	'b']
  for i in illegals:
    if i in name:
      name=name.replace(i,'_')
  return name


class Statistics:
  def __init__(self,ll,panel):
    ll.standardize(panel)
    self.df=panel.df
    self.N,self.T,self.k=panel.X.shape
    self.Rsq_st, self.Rsqadj_st, self.LL_ratio,self.LL_ratio_OLS=stat.goodness_of_fit(ll,True,panel)	
    self.Rsq, self.Rsqadj, self.LL_ratio,self.LL_ratio_OLS=stat.goodness_of_fit(ll,False,panel)	
    self.no_ac_prob,self.rhos,self.RSqAC=stat.breusch_godfrey_test(panel,ll,10)
    self.DW=stat.DurbinWatson(panel,ll)
    self.norm_prob=stat.JB_normality_test(ll.e_norm,panel)
    self.ADF_stat,self.c1,self.c5=stat.adf_test(panel,ll,10)
    self.df_str=self.gen_df_str(panel)	
    self.instruments=panel.input.Z_names[1:]
    self.pqdkm=panel.pqdkm

  def gen_df_str(self,panel):
    summary=f"""
                  SAMPLE SIZE SUMMARY:
\tOriginal sample size\t\t:\t{panel.orig_size}
\tSample size after filtering\t\t:\t{panel.NT_before_loss}
\tDegrees of freedom\t\t:\t{panel.df}
\tNumber of IDs\t\t:\t{self.N:,}
\tNumber of dates (maximum)\t\t:\t{self.T}\n"""		

    group_rmv=f"""
                  REMOVED GROUPS BECAUSE OF TOO FEW OBSERVATIONS:
\tObservations per group lost because of
\tA)\tARIMA/GARCH\t:\t{panel.lost_obs}
\tB)\tMin # of obs in user preferences:\t:\t{panel.options.min_group_df.value}
\tMin # observations required (A+B)\t\t:\t{panel.lost_obs+panel.options.min_group_df.value}\n
\tGroups removed
\tA)\tTotal # of groups\t:\t{len(panel.idincl)}
\tB)\t# of groups removed\t:\t{sum(panel.idincl==False)}
\t# of groups remaining (A-B)\t\t:\t{sum(panel.idincl==True)}

\t# of observations removed\t\t:\t{panel.input.X.shape[0]-panel.NT_before_loss}\n"""


    s=f"""
{summary}
{group_rmv}
                  DEGREES OF FREEDOM:
\tA)\tSample size\t:\t{panel.NT_before_loss}
\tB)\tObservations lost to
\t\tGARCH/ARIMA\t:\t{panel.tot_lost_obs}	
\tRandom/Fixed Effects in
\tC)\tMean process\t:\t{panel.number_of_RE_coef}
\tD)\tVariance process\t:\t{panel.number_of_RE_coef_in_variance}
\tE)\tNumber of coefficients in
\t\tRegression\t:\t{panel.args.n_args:,}
\tDegrees of freedom (A-B-C-D-E)\t\t:\t{panel.df}\n\n"""
    return s

  def gen_mod_fit(self,n_digits = 3):
    return f"""	
\tLL-ratio\t\t:\t{round(self.LL_ratio,n_digits)}
\tR-squared (from observed data)\t\t:\t{round(self.Rsq*100,2)}%
\tAdjusted R-squared (from observed data)\t\t:\t{round(self.Rsqadj*100,2)}%
\tR-squared (from normalized data)\t\t:\t{round(self.Rsq_st*100,2)}%
\tAdjusted R-squared (from normalized data)\t\t:\t{round(self.Rsqadj_st*100,2)}%

\t("Normalized data" means that the data is adjusted with the estimated ARIMA-GARCH parameters and random/fixed effects.)

\tDurbin-Watson statistic:\t\t:\t{round(self.DW,2)}
\tBreusch-Godfrey test\t\t:\t{round(self.no_ac_prob*100,n_digits)}% \t(significance, probability of no auto correlation)
\tJarque–Bera test for normality\t\t:\t{round(self.norm_prob*100,n_digits)}% \t(significance, probability of normality)\n
"""		


  def adf_str(self,n_digits = 3):

    if not self.ADF_stat=='NA':
      if self.ADF_stat<self.c1:
        self.ADF_res="Unit root rejected at 1%"
      elif self.ADF_stat<self.c5:
        self.ADF_res="Unit root rejected at 5%"
      else:
        self.ADF_res="Unit root not rejected"		
      adf=f"""
\tAugmented Dicky-Fuller (ADF)        
\t\tTest statistic\t:\t{round(self.ADF_stat,n_digits)}
\t\t1% critical value\t:\t{round(self.c1,n_digits)}
\t\t5% critical value\t:\t{round(self.c5,n_digits)}
\t\tResult\t:\t\t{self.ADF_res}
                          """
    else:
      adf="Unable to calculate ADF"
    if self.df<1:
      s+="""
\tWARNING: All your degrees of freedom (df) has been consumed, so statistics cannot be computed.
\tyou can increase df by for example turning off random/fixed effects """
    return adf


def get_tab_stops(X,f):

  m_len = f.measure("m")
  counter=2*m_len
  tabs=[f"{counter}",'numeric']
  r,c=np.array(X).shape
  for i in range(c):
    t=1
    num_max=0
    for j in range(r):
      s=str(X[j][i])
      if '.' in s:
        a=s.split('.')
        num_max=max((len(a[0]),num_max))
      t=max((f.measure(X[j][i])+(num_max+2)*m_len,t))
    counter+=t
    tabs.extend([f"{counter}",'numeric'])			
  return tabs

l=STANDARD_LENGTH
#python variable name,	length,		is string,  display name,		neg. values,	justification	next tab space		round digits (None=no rounding,-1=set by user)
pr=[
  ['count',		2,			False,		'',					False,			'right', 		2,					None],
                ['names',		None,		True,		'Variable names:',	False,			'right', 		2, 					None],
                ['args',		None,		False,		'Coef:',			True,			'right', 		2, 					-1],
                ['se_robust',	None,		False,		'rob.SE',			True,			'right', 		3, 					-1],
                ['sign_codes',	5,			True,		'',					False,			'left', 		2, 					-1],
                ['dx_norm',		None,		False,		'direction:',		True,			'right', 		2, 					None],
                ['tstat',		2,			False,		't-stat.:',			True,			'right', 		2, 					2],
                ['tsign',		None,		False,		'p-value:',			False,			'right', 		2, 					3],
                ['multicoll',	1,			True,		'',					False,			'left', 		2, 					None],
                ['assco',		20,			True,		'collinear with',	False,			'center', 		2, 					None],
                ['set_to',		6,			True,		'set to',			False,			'center', 		2, 					None],
                ['cause',		50,			True,		'cause',			False,			'right', 		2, 					None]]		


def format_table(X,cols,fmt,heading, mod, stats,tail):
  if fmt=='NORMAL':
    return mod + stats+format_normal(X,[1],cols)+tail
  if fmt=='LATEX':
    return mod + stats+format_latex(X,cols,heading)+tail
  if fmt=='HTML':
    return format_html(X,cols,heading,stats, mod)+tail	
  if fmt=='CONSOLE':
    return mod + stats+format_console(X)+tail


def format_normal(X,add_rows=[],cols=[]):
  p=''
  if 'multicoll' in cols:
    constr_pos=cols.index('multicoll')+1
    p="\t"*constr_pos+"constraints:".center(38)
  p+="\n"
  for i in range(len(X)):
    p+='\n'*(i in add_rows)
    p+='\n'
    for j in range(len(X[0])):
      p+=f'\t{X[i][j]}'

  return p	

def format_console(X):
  p=''
  X = np.array(X)
  dcol = 2
  n,k = X.shape
  colwith = [max(
          [len(s) for s in X[:,i]])
                   for i in range(k)
                           ]
  for i in range(n):
    p+='\n'
    for j in range(k):
      if j == 0:
        p+=f'{X[i][j]}'.ljust(3)
      elif j==1:
        p+=f'{X[i][j]}'.ljust(colwith[j]+dcol)
      else:
        p+=f'{X[i][j]}'.rjust(colwith[j]+dcol)

  return p	

def format_latex(X,cols,heading):
  X=np.array(X,dtype='U128')
  n,k=X.shape
  p="""
\\begin{table}[ht]
\\caption{%s} 
\\centering
\\begin{tabular}{""" %(heading,)
  p+=' c'*k+' }\n\\hline'
  p+='\t'+' &\t'.join(X[0])+'\\\\\n\\hline\\hline'
  for i in range(1,len(X)):
    p+='\t'+ ' &\t'.join(X[i])+'\\\\\n'
  p+="""
\\hline %inserts single line
\\end{tabular}
\\label{table:nonlin} % is used to refer this table in the text
\\end{table}"""
  return p	

def format_html(X,cols,heading,head, mod):
  X=np.array(X,dtype='U128')
  n,k=X.shape
  if head[-1:] == '\n':
    head = head[:-1]
  head = head.replace('\n',"</td></tr>\n<td class='h'>")
  head = head.replace('\t',"</td><td class='h'>")
  mod = mod.replace('\n','<br>')
  spaces = '&nbsp'*4
  p=(f"<br><br><h1>{heading}</h1><br><br>"
           f"<p>{mod}<br>"
        f"<table class='head'></tr><td class='h'>{head}</td></table><br><br>\n\n"
        f"<p><table>\t<tr><th>"
        )
  p += ('\t</th><th>'+spaces).join(X[0])+'</th></tr>\n'
  for i in range(1,len(X)):
    p+='\t<tr><td>'+'\t</td><td>'.join(X[i])+'</td></tr>\n'
  p+='</table></p>'
  return p		

alphabet='abcdefghijklmnopqrstuvwxyz'
class join_table(dict):
  """Creates a  joint table of several regressions with columns of the join_table_column class.
  See join_table_column for data handling."""
  def __init__(self,args,panel,varnames=[]):
    dict.__init__(self)
    self.names_category_list=list([list(i) for i in args.names_category_list])#making a copy
    k=0
    for i in varnames:
      if i in self.names_category_list[0]:
        k=self.names_category_list[0].index(i)+1
      else:
        self.names_category_list[0].insert(k,i)
        k+=1
    self.caption_v=[itm for s in self.names_category_list for itm in s]#flattening

  def update(self,ll,stats,desc,panel):
    if not desc in self:
      for i in range(len(ll.args.names_category_list)):
        for j in ll.args.names_category_list[i]:
          if not j in self.names_category_list[i]:
            self.names_category_list[i].append(j)
      self.caption_v=[itm for s in self.names_category_list for itm in s]#flattening
    self[desc]=join_table_column(stats, ll,panel)


  def make_table(self, stacked, brackets,digits,caption):
    keys=list(self.keys())
    k=len(keys)
    n=len(self.caption_v)
    if stacked:
      X=[['' for j in range(2+k)] for i in range(4+3*n)]
      for i in range(n):
        X[3*i+1][1]=self.caption_v[i]		
        X[3*i+1][0]=i
      X[1+3*n][1]='Log likelihood'
      X[2+3*n][1]='Degrees of freedom'	
      X[3+3*n][1]='Adjusted R-squared'	
    else:
      X=[['' for j in range(2+2*k)] for i in range(4+n)]
      for i in range(n):
        X[i+1][1]=self.caption_v[i]	
        X[i+1][0]=i
      X[1+n][1]='Log likelihood'
      X[2+n][1]='Degrees of freedom'		
      X[3+n][1]='Adjusted R-squared'	
    for i in range(k):
      self.make_column(i,keys[i],X,stacked, brackets,digits,caption)
    s=format_normal(X,[1,(1+stacked*2)*n+1,(1+stacked*2)*n+4])
    s+=f"\n\nSignificance codes: '=0.1, *=0.05, **=0.01, ***=0.001,    |=collinear\n"
    max_mod=0
    models=[]
    for i in range(len(keys)):
      key=self[keys[i]]
      p,q,d,k,m=key.pqdkm
      models.append(f"\n{alphabet[i]}: {keys[i]}")
      max_mod=max(len(models[i]),max_mod)
    for i in range(len(keys)):
      s+=models[i].ljust(max_mod+2)
      if len(key.instruments):
        s+=f"\tInstruments: {key.instruments}"
      s+=f"\tARIMA({p},{d},{q})-GARCH({k},{m})"
    return s,X

  def make_column(self,col,key,X,stacked, brackets,digits,caption):
    if not 'se_robust' in self[key].stats:
      return

    if caption=='JOINED LONG':
      X[0][(2-stacked)*col+2]+=f"{self[key].Y_name} ({alphabet[col]})"
    else:
      X[0][(2-stacked)*col+2]=alphabet[col]
    n=len(self.caption_v)
    m=len(self[key].args.caption_v)
    ix=[self.caption_v.index(i) for i in self[key].args.caption_v]
    se=np.round(self[key].stats['se_robust'],digits)
    sgn=self[key].stats['sign_codes']
    args=np.round(self[key].args.args_v,digits)
    if brackets=='[':
      se_sgn=[f"[{se[i]}]{sgn[i]}" for i in range(m)]
    elif brackets=='(':
      se_sgn=[f"({se[i]}){sgn[i]}" for i in range(m)]
    else:
      se_sgn=[f"{se[i]}{sgn[i]}" for i in range(m)]				
    if stacked:
      for i in range(m):
        X[3*ix[i]+1][col+2]=args[i]
        X[3*ix[i]+2][col+2]=se_sgn[i]
      X[1+3*n][col+2]=round_sign_digits(self[key].LL,5,1)
      X[2+3*n][col+2]=self[key].df	
      X[3+3*n][col+2]=f"{round(self[key].Rsqadj*100,1)}%"
    else:
      for i in range(m):
        X[ix[i]+1][col*2+2]=args[i]
        X[ix[i]+1][col*2+3]=se_sgn[i]		
      X[1+n][col*2+3]=round_sign_digits(self[key].LL,5,1)
      X[2+n][col*2+3]=self[key].df
      X[3+n][col*2+3]=f"{round(self[key].Rsqadj*100,1)}%"



class join_table_column:
  def __init__(self,stats,ll,panel):
    self.stats=stats
    self.LL=ll.LL
    self.df=panel.df
    self.args=ll.args
    self.Rsq, self.Rsqadj, self.LL_ratio,self.LL_ratio_OLS=stat.goodness_of_fit(ll,True,panel)
    self.instruments=panel.input.Z_names[1:]
    self.pqdkm=panel.pqdkm		
    self.Y_name=panel.input.Y_names





def round_sign_digits(x,digits,min_digits=0):
  d=int(np.log10(abs(x)))
  return np.round(x,max((digits-1,d+min_digits))-d)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from ..output import stat_functions as stat
import numpy as np


def get(g, x, H, constr, f, hessin, ev_constr, simple=True):
  n = len(x)
  if simple or (H is None):
    dx = -(np.dot(hessin,g.reshape(n,1))).flatten()
  else:
    #dx = -(np.dot(hessin,g.reshape(n,1))).flatten()
    #dx, H =solve(self.H, self.g, args.args_v, self.constr, f)
    dx, H, applied_constraints = solve_delete(constr, H, g, x, f, ev_constr)	
  dx_norm = normalize(dx, x)

  return dx, dx_norm, H

def new(g, x, H, constr, f,dx, alam):
  n = len(x)
  dx, slope, rev = slope_check(g, dx)
  if constr is None:
    return dx*alam, slope, rev, []
  elif len(constr.intervals)==0:
    return dx*alam, slope, rev, []
  elif constr.within(x + dx):
    return dx*alam, slope, rev, []

  dxalam, H, applied_constraints = solve_delete(constr, H, g*alam, x, f, None)	
  if np.sum(g*dxalam)<0.0:
    dxalam, H, applied_constraints = solve_delete(constr, H, -g*alam, x, f, None)	
    slope = np.sum(g*dx)
    rev = True
  return dxalam, slope, rev, applied_constraints

def slope_check(g, dx):
  rev = False
  slope=np.sum(g*dx)					#Scale if attempted step is too big.
  if slope <= 0.0:
    dx=-dx
    slope=np.sum(g*dx)
    rev = True  
  return dx, slope, rev


def solve(H, g, x, constr, f, ev_constr):
  """Solves a second degree taylor expansion for the dc for df/dc=0 if f is quadratic, given gradient
  g, hessian H, inequalty constraints c and equalitiy constraints c_eq and returns the solution and 
  and index constrained indicating the constrained variables"""
  if H is None:
    raise RuntimeError('Hessian is None')


  dx_init, H = linalg_solve(H, g, f, x, ev_constr)
  if constr is None:
    return dx_init	
  n=len(H)
  k=len(constr)
  H=np.concatenate((H,np.zeros((n,k))),1)
  H=np.concatenate((H,np.zeros((k,n+k))),0)
  g=np.append(g,(k)*[0])

  for i in range(k):
    H[n+i,n+i]=1
  j=0
  dx=np.zeros(len(g))
  for i in constr.fixed:
    #kuhn_tucker(constr.fixed[i],i,j,n, H, g, x,dx, recalc=False)
    kuhn_tucker2(constr.fixed[i],i,j,n, H, g, x, f,dx, dx_init, ev_constr, recalc=False)
    j+=1
  dx, H = linalg_solve(H, g, f, x, ev_constr)	
  OK=False
  w=0
  for r in range(50):
    j2=j

    for i in constr.intervals:
      dx, H=kuhn_tucker(constr.intervals[i],i,j2,n, H, g, x, f,dx, ev_constr)
      j2+=1
    OK=constr.within(x+dx[:n],False)
    if OK: 
      break
    if r==k+3:
      #print('Unable to set constraints in computation calculation')
      break

  return dx[:n], H


def solve_delete(constr,H, g, x, f, ev_constr):
  """Solves a second degree taylor expansion for the dc for df/dc=0 if f is quadratic, given gradient
  g, hessian H, inequalty constraints c and equalitiy constraints c_eq and returns the solution and 
  and index constrained indicating the constrained variables"""

  if H is None:
    raise RuntimeError('Cant solve with no coefficient matrix')
  try:
    list(constr.keys())[0]
  except:
    dx, H = linalg_solve(H, g, f, x, ev_constr)
    return dx, H
  H = make_hess_posdef(H)
  H_orig = np.array(H)
  m=len(H)

  idx=np.ones(m,dtype=bool)
  delmap=np.arange(m)
  if len(list(constr.fixed.keys()))>0:#removing fixed constraints from the matrix
    idx[list(constr.fixed.keys())]=False
    H=H[idx][:,idx]
    g=g[idx]
    delmap-=np.cumsum(idx==False)
    delmap[idx==False]=m#if for some odd reason, the deleted variables are referenced later, an out-of-bounds error is thrown
  n=len(H)
  k=len(constr.intervals)
  H=np.concatenate((H,np.zeros((n,k))),1)
  H=np.concatenate((H,np.zeros((k,n+k))),0)
  g=np.append(g,(k)*[0])

  for i in range(k):
    H[n+i,n+i]=1
  
  dx, H = linalg_solve(H, g, f, x, ev_constr)
  xi_full=np.zeros(m)
  OK=False
  keys=list(constr.intervals.keys())
  applied_constraints=[]
  for j in range(k):
    key=keys[j]
    dx, H =kuhn_tucker_del(constr,key,j,n, H, g, x, f,dx,delmap, ev_constr)
    applied_constraints.append(key)
    xi_full[idx]=dx[:n]
    OK=constr.within(x+xi_full,False)
    if OK: 
      break

  xi_full=np.zeros(m)
  xi_full[idx]=dx[:n]
  H_full = np.zeros((m,m))
  idx = idx.reshape((1,m))
  nz = np.nonzero(idx*idx.T)
  H_full[nz] = H[np.nonzero(np.ones((n,n)))]
  H_full[H_full==0] = H_orig[H_full==0]
  return xi_full, H_full, applied_constraints


def kuhn_tucker_del(constr,key,j,n,H,g,x, f, dx,delmap, ev_constr,recalc=True):
  q=None
  c=constr.intervals[key]
  i=delmap[key]
  if not c.value is None:
    q=-(c.value-x[key])
  elif x[key]+dx[i]<c.min:
    q=-(c.min-x[key])
  elif x[key]+dx[i]>c.max:
    q=-(c.max-x[key])
  if q!=None:
    H[i,n+j]=1
    H[n+j,i]=1
    H[n+j,n+j]=0
    g[n+j]=q
    if recalc:
      dx, H = linalg_solve(H, g, f, x, ev_constr)
  return dx, H


def kuhn_tucker(c,i,j,n,H,g,x,f,dx, ev_constr,recalc=True):
  q=None
  if not c.value is None:
    q=-(c.value-x[i])
  elif x[i]+dx[i]<c.min:
    q=-(c.min-x[i])
  elif x[i]+dx[i]>c.max:
    q=-(c.max-x[i])
  if q!=None:
    H[i,n+j]=1
    H[n+j,i]=1
    H[n+j,n+j]=0
    g[n+j]=q
    if recalc:
      dx, H = linalg_solve(H, g, f, x, ev_constr)
  return dx, H


def kuhn_tucker2(c,i,j,n,H,g,x, f,dx,dx_init, ev_constr,recalc=True):
  if c.assco_ix is None:
    return kuhn_tucker(c,i,j,n,H,g,x, f, dx,ev_constr, recalc)
  q=None
  if not c.value is None:
    q=-(c.value-x[i])
  elif x[i]+dx[i]<c.min:
    q=-(c.min-x[i])
  elif x[i]+dx[i]>c.max:
    q=-(c.max-x[i])
  if q!=None:
    H[i,n+j]=1
    H[n+j,i]=1
    H[n+j,n+j]=0
    g[n+j]=q
    if recalc:
      dx, H = linalg_solve(H, g, f, x, ev_constr)
  return dx, H




def normalize(dx, x):
  x = np.abs(x)
  dx_norm=(x!=0)*dx/(x+(x<1e-100))
  dx_norm=(x<1e-2)*dx+(x>=1e-2)*dx_norm	
  return dx_norm	



def ev_non_sing(H, ev_constr):
  c, var_prop, ev, p = stat.var_decomposition(XXNorm=H)
  if not ev_constr:
    return ev,p
  c = c.flatten()
  p = p.real
  limit = 100
  v = np.sum(var_prop>0.5,1)<2
  keep = (v|(c<limit))
  ev = ev * keep
  return ev, p

def linalg_solve(H,g, f, x, ev_constr):
  try:
    ev, p = ev_non_sing(H, ev_constr)
    return -np.linalg.solve(H, g), H
  except:
    pass
  #Needs to evaluate this attempt to create a nicer H
  ev, p = ev_non_sing(H, ev_constr)
  a = np.dot(p, p.T)
  b = np.dot(p.T, p)

  gp = np.dot(g, p)

  negev = (ev<0)+(ev==1.0)
  ev = ev*negev - ev*(negev==False)
  ev_1 = 1/(ev + (ev==0))

  gpL = gp*ev_1*(ev!=0)

  dx = -np.real(np.dot(gpL, p.T))

  H = np.dot(p, np.dot(np.diag(ev), p.T))


  return dx, H

def make_hess_posdef(H):
  c, var_prop, ev, p = stat.var_decomposition(XXNorm=H)
  ev = ((ev<0) - 1.0 * (ev>=0)) * ev
  H = np.dot(p, np.dot(np.diag(ev), p.T))  
  return H

def linalg_solve_old(H,g):
  try:
    return -np.linalg.solve(H,g).flatten(), H
  except np.linalg.LinAlgError as e:
    pass
  Hd = np.diag(H)
  Hd = Hd*(np.abs(Hd)>1e-30)
  Hd[Hd==0] = np.min(Hd[Hd<0])
  Hd = np.diag(Hd)
  H = H + Hd
  try:
    return -np.linalg.solve(H,g).flatten(), H
  except np.linalg.LinAlgError as e:
    pass		

  H = H - np.abs(Hd)
  try:
    return -np.linalg.solve(H,g).flatten(), H
  except np.linalg.LinAlgError as e:
    pass

  H =  - np.abs(Hd)
  try:
    return -np.linalg.solve(H,g).flatten(), H
  except np.linalg.LinAlgError as e:
    raise RuntimeError('Unable to fix singularity')		

  H = np.identity(len(Hd))
  return -np.linalg.solve(H,g).flatten(), H



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module contains statistical procedures

from .. import random_effects
from . import stat_dist
from .. import functions as fu
import numpy as np



def var_decomposition(XXNorm=None,X=None):
  """Variance decomposition. Returns the matrix of condition indexes for each factor (rows) and each variable
  (columns). Calculates the normalized sum of squares using square_and_norm if XXNorm is not supplied"""
  if XXNorm is None:
    XXNorm=square_and_norm(X)
  ub=len(XXNorm)     
  ev,p=np.linalg.eig(XXNorm)
  if np.any(np.round(ev.imag,15)!=len(ev)*[0]):
    pass
    #print( "non-real XX matrix")

  ev=ev.real
  p=p.real
  d=np.abs(ev)**0.5+1e-100
  MaxEv=np.max(d)  
  fi=np.abs(p*p/((d*d).reshape((1,ub))+1E-200))
  fiTot=np.sum(fi,1).reshape((ub,1))
  pi=fi/(fiTot + (fiTot==0)*1E-200)
  pi=pi.T
  CondIx=MaxEv/d
  ind=np.argsort(CondIx)
  pi=pi[ind]
  CondIx=CondIx[ind]
  CondIx=CondIx.reshape((len(CondIx),1))
  return CondIx,pi, ev, p

def square_and_norm(X):
  """Squares X, and normalize to unit lenght.
  Similar to a correlation matrix, except the
  means are not subtracted"""
  N,T,k=X.shape
  Sumsq=np.sqrt(np.sum(np.sum(X**2,0),0))
  Sumsq.resize((k,1))
  Sumsq=Sumsq*Sumsq.T
  norm=fu.dot(X,X)/(Sumsq+1e-200)
  return norm

def singular_elim(panel,X):
  """Eliminates variables that cause singularity"""
  N,T,k=X.shape
  r=np.arange(k)
  ci_threshold=50
  keep,XXCorrel=find_singulars(panel,X) 
  XXNorm=square_and_norm(X)
  cond_ix, pi, d, p =var_decomposition(XXNorm)
  if max(cond_ix)<ci_threshold:
    return keep,cond_ix
  for cix in range(1,len(cond_ix)):
    if (np.sum(pi[-cix]>0.5)>1) and cond_ix[-cix][0]>ci_threshold:
      keep[pi[:,-cix]>0.5]=False
  return keep,cond_ix

def find_singulars(panel,X):
  """Returns a list with True for variables that cause singularity and False otherwise.
  for the main regression, singularity is detected by the constraints module"""
  N,T,k=X.shape
  XXCorrel=correl(X,panel)
  keep=np.all(np.isnan(XXCorrel)==False,0)
  keep=keep*np.all((np.abs(np.triu(XXCorrel,1))>0.99)==False,0)
  x_dev=deviation(panel, X)
  var_x=np.sum(np.sum(x_dev**2,0),0)
  keep=keep*(var_x>0)#remove constants
  keep[0]=True#allways keep the first constant term
  return keep,XXCorrel

def adf_test(panel,ll,p):
  """Returns the augmented dickey fuller test statistic and critical value"""
  N,T,k=panel.X.shape
  y=ll.Y_st
  yl1=roll(y,1,1)
  dy=y-yl1
  date_var=np.arange(T).reshape((T,1))*panel.included[3]	#date count
  X=np.concatenate((panel.included[3],date_var,yl1),2)
  dyL=[]
  for i in range(p):
    dyL.append(roll(dy,i+1,1))
  dyL=np.concatenate(dyL,2)
  date_var=(date_var>p+1)
  X=np.concatenate((X,dyL),2)
  X=X*date_var
  dy=dy*date_var
  X[:,0:panel.lost_obs+10]=0
  keep,c_ix=singular_elim(panel,X)
  if not np.all(keep[0:3]):
    return 'NA','NA','NA'
  beta,se_robust,se=OLS(panel,X[:,:,keep],dy,robust_se_lags=10,c=date_var)
  adf_stat=beta[2]/se_robust[2]
  critval=adf_crit_values(panel.NT,True)
  res=np.append(adf_stat,critval)
  return res

def roll(x,shift, dim):
  b = np.roll(x,shift,dim)
  b[:,:shift]=0
  return b

def goodness_of_fit(ll,standarized,panel):
  if standarized:
    s_res=panel.var(ll.e_RE)
    s_tot=panel.var(ll.Y_st)
  else:
    s_res=panel.var(ll.u)
    s_tot=panel.var(panel.Y)		
  r_unexpl=s_res/s_tot
  Rsq=1-r_unexpl
  Rsqadj=1-r_unexpl*(panel.NT-1)/(panel.NT-panel.args.n_args-1)
  panel.args.create_null_ll(panel)
  LL_ratio_OLS=2*(ll.LL-panel.args.LL_OLS)
  LL_ratio=2*(ll.LL-panel.args.LL_null)
  return Rsq, Rsqadj, LL_ratio,LL_ratio_OLS


def DurbinWatson(panel,ll):
  """returns the probability that err_vec are not auto correlated""" 
  X=ll.XIV_st
  N,T,k=X.shape	
  e=ll.e_norm_centered
  c=panel.included[3][:,:-1]
  DW=np.sum((c*(e[:,1:]-e[:,:-1]))**2)/np.sum((e*panel.included[3])**2)
  return DW

def correlogram(panel,e,lags,center=False):
  N,T,k=e.shape	
  v=panel.var(e)
  if center:
    e=e-panel.mean(e)
  rho=np.zeros(lags+1)
  rho[0]=1
  for i in range(1,lags+1):
    a=panel.T_i-i-1>0
    incl=(a*panel.included[3])[:,i:,:]
    df=np.sum(incl)
    rho[i]=np.sum(incl*e[:,i:]*e[:,0:-i])/(v*df)
  return rho #The probability of no AC given H0 of AC.



def adf_crit_values(n,trend):
  """Returns 1 and 5 percent critical values respectively. Interpolated from table in 
  Fuller, W. A. (1976). Introduction to Statistical Time Series. New York: John Wiley and Sons. ISBN 0-471-28715-6
  table is available at https://en.wikipedia.org/wiki/Augmented_Dickey%E2%80%93Fuller_test"""
  if trend:
    d={25:np.array([-3.75,-3.00]),50:np.array([-3.58,-2.93]),100:np.array([-3.51,-2.89]),
                   250:np.array([-3.46,-2.88]),500:np.array([-3.44,-2.87]),10000:np.array([-3.43,-2.86])}
  else:
    d={25:np.array([-4.38,-3.60]),50:np.array([-4.15,-3.50]),100:np.array([-4.04,-3.45]),
                   250:np.array([-3.99,-3.43]),500:np.array([-3.98,-3.42]),10000:np.array([-3.96,-3.41])}

  if n<25:
    print ("Warning: ADF critical values are not available for fewer than 25 observations")
    return (0,0)
  k=(25,50,100,250,500,10000)
  r=None
  for i in range(len(k)-1):
    if n>=k[i] and n<500:
      r=d[k[i]]+(n-k[i])*((d[k[i+1]]-d[k[i]])/(k[i+1]-k[i]))#interpolation
      return r
  if r is None:
    return d[10000]


def correl(X,panel=None, covar=False):
  """Returns the correlation of X. Assumes three dimensional matrices. """
  if not panel is None:
    X=X*panel.included[3]
    N,T,k=X.shape
    N=panel.NT
    mean=np.sum(np.sum(X,0),0).reshape((1,k))/N
  else:
    N,k=X.shape
    mean=np.sum(X,0).reshape((1,k))/N
  cov=fu.dot(X,X)/N

  cov=cov-(mean.T*mean)

  if covar:
    return cov

  stdx=(np.diag(cov)**0.5).reshape((1,k))
  stdx=(stdx.T*stdx)
  stdx[np.isnan(stdx)]=0
  corr=(stdx>0)*cov/(stdx+(stdx==0)*1e-100)
  corr[stdx<=0]=0

  return corr

def deviation(panel,X):
  N,T,k=X.shape
  x=X*panel.included[3]
  mean=np.sum(np.sum(x,0),0).reshape((1,1,k))/panel.NT
  return (X-mean)*panel.included[3]

def correl_2dim(X,Y=None,covar=False):
  """Returns the correlation of X and Y. Assumes two dimensional matrixes. If Y is not supplied, the 
  correlation matrix of X is returned"""	
  if type(X)==list:
    X=Concat(X)
  single=Y is None
  if single:
    Y=X
  T,k=X.shape
  T,m=Y.shape
  X_dev=X-np.mean(X,0)
  Y_dev=Y-np.mean(Y,0)
  cov=np.dot(X_dev.T,Y_dev)
  if covar:
    return cov/(len(X)-1)
  stdx=np.sum(X_dev**2,0).reshape((1,k))**0.5
  if single:
    stdy=stdx
  else:
    stdy=np.sum(Y_dev**2,0).reshape((1,k))**0.5
  std_matr=stdx.T*stdy
  std_matr=std_matr+(std_matr==0)*1e-200
  corr=cov/std_matr
  if corr.shape==(1,1): 
    corr=corr[0][0]
  return corr

def get_singular_list(panel,XX):
  a,b=singular_elim(panel,XX)
  names=np.array(panel.input.X_names)[a==False]
  idx=np.array(range(len(a)))[a==False]
  s=', '.join([f"{names[i]}" for i in range(len(idx))])	
  return s

def OLS(panel,X,Y,add_const=False,return_rsq=False,return_e=False,c=None,robust_se_lags=0):
  """runs OLS after adding const as the last variable"""
  if c is None:
    c=panel.included[3]
  N,T,k=X.shape
  NT=panel.NT
  if add_const:
    X=np.concatenate((c,X),2)
    k=k+1
  X=X*c
  Y=Y*c
  XX=fu.dot(X,X)
  XY=fu.dot(X,Y)
  try:
    beta=np.linalg.solve(XX,XY)
  except np.linalg.LinAlgError:
    s=get_singular_list(panel,X)
    raise RuntimeError("The following variables caused singularity runtime and must be removed: "+s)
  if return_rsq or return_e or robust_se_lags:
    e=(Y-fu.dot(X,beta))*c
    if return_rsq:
      v0=panel.var(e,included=c)
      v1=panel.var(Y,included=c)
      Rsq=1-v0/v1
      #Rsqadj=1-(v0/v1)*(NT-1)/(NT-k-1)
      return beta,Rsq
    elif return_e:
      return beta,e*c
    elif robust_se_lags:
      XXInv=np.linalg.inv(XX)
      se_robust,se,V=robust_se(panel,robust_se_lags,XXInv,X*e)
      return beta,se_robust.reshape(k,1),se.reshape(k,1)
  return beta

def OLS_simple(Y,X,addconst=False,residuals=True):
  """Returns the OLS residuals if residuals. For use with two dimiensional arrays only"""
  if addconst:
    n=len(X)
    X=np.concatenate((np.ones((n,1)),X),1)
  XY=np.dot(X.T,Y)
  XX=np.dot(X.T,X)
  XXInv=np.linalg.inv(XX)
  beta=np.dot(XXInv,XY)
  e=Y-np.dot(X,beta)
  if residuals:
    return beta,e
  else:
    return beta

def newey_west_wghts(L,XErr):
  """Calculates the Newey-West autocorrelation consistent weighting matrix. Either err_vec or XErr is required"""
  N,T,k=XErr.shape
  S=np.zeros((k,k))
  try:
    a=min(L,T)
  except:
    a=0
  for i in range(1,min(L,T)):
    w=1-(i+1)/(L)
    XX=fu.dot(XErr[:,i:],XErr[:,0:T-i])
    S+=w*(XX+XX.T)
  return S

def robust_cluster_weights(panel,XErr,cluster_dim,whites):
  """Calculates the Newey-West autocorrelation consistent weighting matrix. Either err_vec or XErr is required"""
  N,T,k=XErr.shape
  if cluster_dim==0:#group cluster
    if N<=1:
      return 0
    mean=panel.mean(XErr,0)
  elif cluster_dim==1:#time cluster
    mean=random_effects.mean_time(panel,XErr,True)
    T,m,k=mean.shape
    mean=mean.reshape((T,k))
  S=fu.dot(mean,mean)-whites
  return S


def robust_se(panel,L,hessin,XErr,nw_only=True):
  """Returns the maximum robust standard errors considering all combinations of sums of different combinations
  of clusters and newy-west"""
  w,W=sandwich_var(hessin,fu.dot(XErr,XErr))#whites
  nw,NW=sandwich_var(hessin,newey_west_wghts(L,XErr))#newy-west
  if panel.N>1:
    c0,C0=sandwich_var(hessin,robust_cluster_weights(panel,XErr, 0, w))#cluster dim 1
    c1,C1=sandwich_var(hessin,robust_cluster_weights(panel,XErr, 1, w))#cluster dim 2
  else:
    c0,c1,C0,C1=0,0,0*W,0*W
  v=np.array([
          nw,
                nw+c0,
                nw+c1,
                nw+c1+c0,
                w*0
        ])
  V=np.array([
          NW,
                NW+C0,
                NW+C1,
                NW+C1+C0,
                W*0
        ])
  V=V+W
  s=np.max(w+v,0)
  se_robust=np.maximum(s,0)**0.5
  i=np.argmax(np.sum(w+v,1))
  se_std=np.maximum(w,0)**0.5
  return se_robust,se_std,V[i]

def sandwich_var(hessin,V):
  hessinV=np.dot(hessin,V)
  V=np.dot(hessinV,hessin)
  v=np.diag(V)
  return v,V


def breusch_godfrey_test(panel,ll, lags):
  """returns the probability that err_vec are not auto correlated""" 
  e=ll.e_norm_centered
  X=ll.XIV_st
  N,T,k=X.shape
  X_u=X[:,lags:T]
  u=e[:,lags:T]
  c=panel.included[3][:,lags:T]
  for i in range(1,lags+1):
    X_u=np.append(X_u,e[:,lags-i:T-i],2)
  Beta,Rsq = OLS(panel,X_u,u,False,True,c=c)
  T=(panel.NT-k-1-lags)
  BGStat=T*Rsq
  rho=Beta[k:]
  ProbNoAC=1.0-stat_dist.chisq(BGStat,lags)
  return ProbNoAC, rho, Rsq #The probability of no AC given H0 of AC.




def JB_normality_test(e,panel):
  """Jarque-Bera test for normality. 
  returns the probability that a set of residuals are drawn from a normal distribution"""
  e=e[panel.included[3]]
  a=np.argsort(np.abs(e))[::-1]

  ec=e[a][int(0.001*len(e)):]

  df=len(ec)
  ec=ec-np.mean(ec)
  s=(np.sum(ec**2)/df)**0.5
  mu3=np.sum(ec**3)/df
  mu4=np.sum(ec**4)/df
  S=mu3/s**3
  C=mu4/s**4
  JB=df*((S**2)+0.25*(C-3)**2)/6.0
  p=1.0-stat_dist.chisq(JB,2)
  return p

